## STAMP WorkbenchをPowerShellからビルドおよび実行するための手順書
STAMP Workbench をWindowsのPowerShellからビルドおよび実行するための手順を下記に記す。
なお、当手順の動作を確認できる環境は下記の通りである。
- Windows 10 バージョン20H2
- AdoptOpenJDK（または Eclipse Adoptium OpenJDK） バージョン1.8.0_292
- Apache Maven バージョン3.6.3
- Windows PowerShell バージョン5.1

### STAMP Workbenchを用意する
  - [本サイト](https://www.ipa.go.jp/sec/tools/stamp_workbench.html)より、STAMP Workbench ver2.0.0をダウンロードする。
  - サイトの手順に従い、STAMP Workbenchをインストールする。

### STAMP Workbench ソースコードを用意する
  - [本サイト](https://www.ipa.go.jp/sec/tools/stamp_workbench.html)より、『source_code.zip』をダウンロードする。
  - 『source_code.zip』を任意のフォルダに展開する。

### AdoptOpenJDK（または Eclipse Adoptium OpenJDK）をインストールする
- Apache Mavenを動作させるための準備として、JDK(Java Development Kit)をインストールする。
- JDKを展開したフォルダを JAVA_HOME という名前で環境変数に設定する。

### Apache Mavenを導入する
STAMP Workbenchをビルド・実行するために、プロジェクト管理ツールであるMavenをインストールする。下記に、Maven導入のための手順の一例を示す。
  - Mavenをダウンロードする。
     - <http://maven.apache.org/download.html>
  - ダウンロードしたファイルを任意のフォルダに展開する。
 - Maven を展開したフォルダを M2_HOME という名前で環境変数に設定する
 - PATH 環境変数に ${M2_HOME}\bin を追加する
 - 最後に、PowerShellを開き、`mvn -v`を実行して、Mavenが実行できることを確認する。

### ソースコードをPowerShellからビルドする
  - PowerShellを起動し、ダウンロードしたSTAMP Workbenchのソースコードを展開したフォルダ配下の `stamp-project\projects` へ移動する。
  - ビルドのコマンドを実行するための事前準備として、PowerShellから変数の設定を行う。
    - `$stampworkbench = "STAMP Workbench をインストールしたフォルダのパスを指定する"`
    - `$any_path = "任意のフォルダのパスを指定する。当フォルダはMavenが依存関係のダウンロードやビルド済みのjarを配置するために使用する"`
  - 下記のコマンドを実行する。
     - `mvn "-Dstampworkbench=${stampworkbench}" "-Dmaven.repo.local=${any_path}" -s "${env:M2_HOME}/conf/settings.xml" clean install`
 - コマンドを実行し、下記のようなログが表示されればビルド成功。
 
``` 
[INFO] ------------------------------------------------------------------------  
[INFO] Reactor Summary:  
[INFO]  
[INFO] STAMP Workbench - Root POM ......................... SUCCESS [  0.629 s]  
[INFO] STAMP Workbench - STAMP/STPA Metamodel ............. SUCCESS [ 10.291 s]  
[INFO] STAMP Workbench - STAMP/STPA Notation Metamodel .... SUCCESS [  5.644 s]  
[INFO] STAMP Workbench - STAMP/STPA Metamodel Edit ........ SUCCESS [  5.051 s]  
[INFO] STAMP Workbench - STAMP/STPA Notation Metamodel Edit SUCCESS [  4.688 s]  
[INFO] STAMP Workbench - STAMP/STPA Notation Editor ....... SUCCESS [  3.728 s]  
[INFO] STAMP Workbench - UI ............................... SUCCESS [  9.307 s]  
[INFO] STAMP Workbench - Application Bootstrap ............ SUCCESS [  2.204 s]  
[INFO] ------------------------------------------------------------------------  
[INFO] BUILD SUCCESS  
[INFO] ------------------------------------------------------------------------  
```


### PowerShell から STAMP Workbench を起動する
  - 下記のコマンドを実行し、STAMP Workbenchを起動する。
    - `mvn "-Dstampworkbench=${stampworkbench}" "-Dmaven.repo.local=${any_path}" exec:java -f net.astah.stpa.stamp.app/pom.xml`
 - STAMP Workbench が起動できたら成功。

### Maven に依存せずに起動する方法
  - STAMP Workbench のフォルダ内にビルド結果の JAR を配置することで Maven に依存せず起動することが可能です。
  - 各プロジェクトのビルド成果物 (`*/target/*.jar`) で STAMP Workbench の同名の JAR を置き換える形になります。
    - ただし、 `stampworkbench.jar` は `net.astah.stpa.stamp.app/target/*.jar` で名前が一致しません。
  - 起動用の `stampworkbench.jar` は `META-INF/MANIFEST.MF` の差し替えが必要になります。
    - STAMP Workbench に同梱されている JAR から上記ファイルを抽出してビルド結果の JAR を更新します。
    - JAR の更新は `jar umf ${抽出した MANIFEST.MF} stampworkbench.jar -C ${空のディレクトリ} .` 等で行えます。
