package net.astah.stpa.stamp;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;

import net.astah.dd.dc.constraints.DCDelegatingConstraintProvider;
import net.astah.dd.dg.constraints.DGDelegatingConstraintProvider;
import net.astah.dd.di.constraints.DIConstraintProvider;
import net.astah.emf.edit.job.CommandJob;
import net.astah.emf.edit.job.Jobs;
import net.astah.emf.validation.ModeledConstraintsHelper;
import net.astah.golf.control.CommandEvent;
import net.astah.golf.control.CommandExecutionListener;
import net.astah.golf.control.CommandManager;
import net.astah.jmodel.ProjectManager;
import net.astah.jomt.Amerigo;
import net.astah.jomt.jsystem.ModelGlobal;
import net.astah.jomt.jutil.SwingInvoker;
import net.astah.jomt.jview.MessageManager;
import net.astah.logger.LoggerInitialization;
import net.astah.stpa.stamp.constraints.STAMPConstraintProvider;
import net.astah.stpa.stamp.notation.provider.STAMPNotationEditPluginHelper;
import net.astah.stpa.stamp.provider.STAMPEditPluginHelper;

/**
 * STAMP/STPA モデルエディタを起動するためのメインクラスです。
 */
public class STPA {

    static {
        LoggerInitialization.initialize(STPA.class);
    }

    /**
     * Java 仮想マシンから起動時に呼び出されるメソッドです。
     * 
     * @param args
     *            Java 仮想マシンの起動時に指定されたコマンドライン引数の配列
     */
    public static void main(final String args[]) {
        CommandManager.instance().setCommandExecutionListener(new CommandExecutionListener() {
            @Override
            public void commandExecuted(final CommandEvent event) {
            }

            @Override
            public void commandAborted(final CommandEvent event, final Throwable throwable) {
                MessageManager.requireBugReport(throwable);
            }
        });
        Jobs.blockEDT(Job.INTERACTIVE);
        Job.getJobManager().addJobChangeListener(new JobChangeAdapter() {
            @Override
            public void done(final IJobChangeEvent event) {
                if (!(event.getJob() instanceof CommandJob)) {
                    return;
                }

                final IStatus result = event.getResult();
                if ((result != null) && (result.getSeverity() == IStatus.ERROR)) {
                    final Throwable exception = result.getException();
                    if ((exception instanceof RuntimeException) || (exception instanceof Error)) {
                        SwingInvoker.runInEDT(() -> {
                            MessageManager.requireBugReport(exception);
                        });
                    }
                }
            }
        });

        ProjectManager prjManager = ModelGlobal.prjManager;
        if (prjManager == null) {
            prjManager = ProjectManager.instance();
        }
        prjManager.addPropertyChangeListener(new PropertyChangeListener() {
            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                if ("isLoadedPrj".equals(evt.getPropertyName()) && (Boolean) evt.getNewValue()) {
                    new ModeledConstraintsHelper().loadConstraintBundles(
                            DIConstraintProvider.class.getResource("di.validation"),
                            DCDelegatingConstraintProvider.class.getResource("dc.validation"),
                            DGDelegatingConstraintProvider.class.getResource("dg.validation"),
                            STAMPConstraintProvider.class.getResource("stamp.validation"),
                            getClass().getResource("/net/astah/constraints/bindings.validation"));
                    ((ProjectManager) evt.getSource()).removePropertyChangeListener(this);
                }
            }
        });

        STAMPNotationEditPluginHelper.init();
        STAMPEditPluginHelper.init();
        Amerigo.main(args);
    }

}
