package net.astah.stpa.stamp.provider;

import static java.util.Arrays.asList;
import static java.util.Collections.singletonList;

import java.util.Comparator;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;

import net.astah.emf.common.ecore.EComparators;
import net.astah.emf.common.notify.EFeaturePath;
import net.astah.emf.edit.provider.ContentDecorator;
import net.astah.emf.edit.provider.ContentDecorator.Registry;
import net.astah.emf.edit.provider.decorator.BasicContentDecoratorFactoryImpl;
import net.astah.emf.edit.provider.decorator.FilterRule;
import net.astah.emf.edit.provider.decorator.MoveClassRule;
import net.astah.emf.edit.provider.decorator.SortRule;
import net.astah.stpa.stamp.NamedElement;
import net.astah.stpa.stamp.STAMPPackage;

/**
 * STAMP/STPA モデルの構造ツリー向けの修飾ルールを構築するクラスです。
 */
public class STAMPTreeDecoratorFactory extends BasicContentDecoratorFactoryImpl {

    /**
     * STAMP/STPA モデル要素のソートに使用する比較関数です。
     */
    protected final Comparator<Object> COMPARATOR = STAMPTreeComparator.INSTANCE.thenComparing(
            EComparators.cast(NamedElement.class, Comparator.comparing(NamedElement::getName,
                    Comparator.nullsLast(String.CASE_INSENSITIVE_ORDER))));

    /**
     * @see BasicContentDecoratorFactoryImpl#addContentDecorators(ContentDecorator.Registry)
     */
    @Override
    protected void addContentDecorators(final Registry registry) {
        super.addContentDecorators(registry);

        // ComponentをSTPAAnalysis配下に表示する
        registry.addDecorator(
                MoveClassRule.moveClass(EFeaturePath.builder()
                        .add(STAMPPackage.Literals.STPA_ANALYSIS__CONTROL_STRUCTURE)
                        .add(STAMPPackage.Literals.CONTROL_STRUCTURE__COMPONENT).create()),
                Priority.ADD.value());

        // STPAAnalysis直下のComponent以外のモデルをフィルタする
        registry.addDecorator(FilterRule.excludeKindOf(STAMPPackage.Literals.STPA_ANALYSIS,
                asList(STAMPPackage.Literals.IDENTIFIED_ELEMENT,
                        STAMPPackage.Literals.UCA_GUIDE_WORD_SET,
                        STAMPPackage.Literals.HCF_HINT_WORD_SET,
                        STAMPPackage.Literals.CONTROL_STRUCTURE,
                        STAMPPackage.Literals.UCA_ANALYSIS,
                        STAMPPackage.Literals.HCF_ANALYSIS)),
                Priority.FILTER.value());
    }

    /**
     * @see BasicContentDecoratorFactoryImpl#addOrderingDecorators(ContentDecorator.Registry)
     */
    @Override
    protected void addOrderingDecorators(final Registry registry) {
        super.addOrderingDecorators(registry);

        // ツリー内でのソート
        registry.addDecorator(SortRule.sort(STAMPPackage.Literals.ELEMENT, COMPARATOR),
                Priority.ORDER.value());
    }

    /**
     * @see BasicContentDecoratorFactoryImpl#prepareExcludeTypeOf()
     */
    @Override
    protected Map<EClass, Set<EClass>> prepareExcludeTypeOf() {
        final Map<EClass, Set<EClass>> excludeTypeOf = super.prepareExcludeTypeOf();

        // プロセスモデルは表示しない
        addValues(excludeTypeOf, STAMPPackage.Literals.COMPONENT,
                singletonList(STAMPPackage.Literals.PROCESS_MODEL));

        return excludeTypeOf;
    }

}
