package net.astah.stpa.stamp.ui.control;

import java.awt.Frame;
import java.beans.PropertyChangeEvent;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.swing.JTable;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.util.EcoreUtil;

import com.google.common.collect.Iterators;
import com.google.common.collect.Lists;

import net.astah.dd.di.Diagram;
import net.astah.golf.control.Command;
import net.astah.golf.model.IEntityStore;
import net.astah.golf.widget.CDesktopPane;
import net.astah.golf.widget.swing.SwingComponent;
import net.astah.jmodel.Project;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jsystem.ModelGlobal;
import net.astah.jomt.jutil.SaveFileChooser;
import net.astah.jomt.jview.MessageManager;
import net.astah.jomt.jview.TableEditor;
import net.astah.notation.NotationPackage;
import net.astah.stpa.stamp.notation.provider.STAMPNotationTreeComparator;
import net.astah.stpa.stamp.ui.Messages;
import net.astah.table.ui.export.TableExcelExport;

/**
 * 表をまとめて Excel に出力するコマンドです。
 */
public class AllExportExcelCommand extends Command {

    @Override
    public void execute() {
        final List<Diagram> diagrams = getOutputDiagrams();
        if (diagrams.isEmpty()) {
            MessageManager.displayInfoMessageDialog("", Messages.all_export_excel_none_warning);
            return;
        }
        final Frame parentCmp = (Frame) ((SwingComponent) JomtGlobal.viewManager.getManagementView()
                .getComponent()).getComponent();
        new SaveFileChooser(parentCmp, "xlsx", "xlsx Files", null, false).select(file -> {
            final List<JTable> tables = new ArrayList<>();
            final List<TableEditor> tableEditors = new ArrayList<>();
            for (Diagram diagram : diagrams) {
                final TableEditor tableEditor = createMatrixEditor(diagram);
                final JTable table = (JTable) tableEditor.getTable().getComponent();
                table.setName(diagram.getName());
                tables.add(table);
                tableEditors.add(tableEditor);
            }

            new TableExcelExport(tables).export(file.getPath());
            tableEditors.forEach(m -> m.remove());
            return true;
        });
    }

    private List<Diagram> getOutputDiagrams() {
        Project currentProject = ModelGlobal.prjManager.getCurrentProject();
        if (currentProject == null) {
            return Collections.emptyList();
        }
        IEntityStore doc = currentProject.doc;
        ResourceSet resourceSet = doc.getDomain().getResourceSet();
        TreeIterator<Object> iterator = EcoreUtil.getAllContents(resourceSet, false);

        @SuppressWarnings("unchecked")
        List<Diagram> diagrams = (List<Diagram>) (Object) Lists.newArrayList(
                Iterators.filter(iterator, NotationPackage.Literals.TABLE_VIEW::isInstance));
        diagrams.sort(new STAMPNotationTreeComparator());

        return diagrams;
    }

    private TableEditor createMatrixEditor(Diagram diagram) {
        CDesktopPane desktopPane = JomtGlobal.viewManager.getManagementView().getDesktopPane();
        TableEditor editor = new TableEditor(diagram, desktopPane) {
            // TableEditor::maximize()を実行した時に図の整列状態が解除されないようにする
            @Override
            public void componentPropertyChange(final PropertyChangeEvent e) {
            }
        };
        editor.maximize();
        return editor;
    }
}
