package net.astah.stpa.stamp.ui.control;

import static java.util.stream.Collectors.toList;

import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.command.IdentityCommand;
import org.eclipse.emf.common.command.StrictCompoundCommand;
import org.eclipse.emf.common.util.ECollections;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CopyCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.collect.Iterables;

import net.astah.dd.dc.Bounds;
import net.astah.dd.dc.DCFactory;
import net.astah.dd.dc.Point;
import net.astah.dd.di.DIPackage;
import net.astah.dd.di.Diagram;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.edit.command.DiagramContext;
import net.astah.dd.di.edit.command.MoveTogetherCommand;
import net.astah.dd.di.edit.command.SetBoundsCommand;
import net.astah.dd.di.edit.provider.Geometries;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.common.ecore.DataTypes;
import net.astah.emf.common.ecore.EComparators;
import net.astah.emf.edit.command.CommandUtils;
import net.astah.emf.edit.command.Commands;
import net.astah.emf.edit.command.DeferringCommandWrapper;
import net.astah.emf.edit.command.PasteCommand;
import net.astah.golf.geom2D.Pnt2d;
import net.astah.golf.geom2D.Rectangle2d;
import net.astah.golf.geom2D.Vec2d;
import net.astah.golf.model.IEntityStore;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jsystem.ModelGlobal;
import net.astah.jomt.jsystem.PropertyManager;
import net.astah.notation.Edge;
import net.astah.notation.LineConstraint;
import net.astah.notation.NotationFactory;
import net.astah.notation.Shape;
import net.astah.notation.Vector;
import net.astah.notation.edit.layout.DiagramElementOperations;
import net.astah.project.Facet;
import net.astah.project.Facets;
import net.astah.project.Project;
import net.astah.project.ProjectManagerAdapter;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.ControlAction;
import net.astah.stpa.stamp.ControlLink;
import net.astah.stpa.stamp.ControlLoop;
import net.astah.stpa.stamp.HCFAnalysis;
import net.astah.stpa.stamp.HCFHintWordSet;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.LinkEnd;
import net.astah.stpa.stamp.STAMPPlugin;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.Signal;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.notation.STAMPControlLoopDiagram;
import net.astah.stpa.stamp.notation.STAMPDiagramElement;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;
import net.astah.stpa.stamp.notation.edit.command.CreateSTAMPNotationRootCommand;
import net.astah.stpa.stamp.notation.edit.command.SynchronizeLabelCommand;
import net.astah.stpa.stamp.notation.provider.STAMPComponentShapeItemProvider;
import net.astah.ui.control.CreateEdgeCommand;
import net.astah.ui.control.CreateShapeCommand;
import net.astah.uml.exception.IllegalModelTypeException;

/**
 * コントロールループ図を作成するコマンドです。
 */
public class CreateControlLoopDiagramCommand extends CreateSTAMPDiagramFromPrjCommand {
    private static final Logger LOGGER = LoggerFactory
            .getLogger(CreateControlLoopDiagramCommand.class);

    private DiagramElement selectedDiagramElement = null;
    private Collection<?> copyResults;
    private final List<DiagramElement> diagramSelections = new ArrayList<>();
    private DiagramElement diagramElement;
    private DiagramElement sourceElement;
    private DiagramElement targetElement;
    private final List<Edge> externalLinks = new ArrayList<>();

    /**
     * 選択されている図要素モデルを設定します。
     * 
     * @param selectedDiagramElement
     *            選択されている図要素モデル
     */
    public void setSelectedDiagramElement(final DiagramElement selectedDiagramElement) {
        this.selectedDiagramElement = selectedDiagramElement;
    }

    /**
     * 選択されている図要素モデルを返します。
     * 
     * @return 選択されている図要素モデル
     */
    public DiagramElement getSelectedDiagramElement() {
        return selectedDiagramElement;
    }

    @Override
    protected Diagram createNewDiagram(final IEntityStore doc, final EObject owner) {
        if (selectedDiagramElement == null) {
            return null;
        }

        if (selectedDiagramElement instanceof STAMPLabel) {
            diagramElement = selectedDiagramElement.getOwningElement();
        } else {
            diagramElement = selectedDiagramElement;
        }

        collectElements();
        if (!copyRelatedDiagramElements()) {
            return null;
        }

        final EditingDomain domain = doc.getDomain();
        final Project project = ProjectManagerAdapter.getRootProject(owner);
        final Facet facet = project.getFacet(STAMPPlugin.FACET_SYMBOLIC_NAME);
        final STPAAnalysis analysis = (STPAAnalysis) facet.getContents().get(0);

        HCFAnalysis hcfAnalysis = analysis.getHcfAnalysis();
        if (hcfAnalysis == null) {
            hcfAnalysis = analysis.createHcfAnalysis();
        }

        // CL図作成対象のControlLinkをtargetLinkに持つControlLoopがすでに存在している場合
        final ControlLoop controlLoop = getControlLoop(hcfAnalysis);
        if (controlLoop == null) {
            return null;
        }
        if (controlLoop.getTargetLink() == null && diagramElement instanceof STAMPEdge) {
            DiagramElements.ifModelElementPresent(diagramElement, ControlLink.class, m -> {
                controlLoop.setTargetLink(m);
            });
        }

        // ヒントワードがセットされていない場合はデフォルトのヒントワードをセットする
        if (controlLoop.getHintWordSet() == null) {
            HCFHintWordSet defaultHintWordSet = analysis.getHCFHintWordSets().get(0);
            controlLoop.setHintWordSet(defaultHintWordSet);
        }

        final STAMPControlLoopDiagram controlLoopDiagram = STAMPNotationFactory.eINSTANCE
                .createSTAMPControlLoopDiagram();
        Commands.execute(new CreateSTAMPNotationRootCommand(domain, facet, owner,
                controlLoopDiagram));
        controlLoopDiagram.setControlLoop(controlLoop);

        return controlLoopDiagram;
    }

    @Override
    protected void setDiagramName(final IEntityStore doc, final Diagram diagram) {
        final ResourceSet resourceSet = doc.getDomain().getResourceSet();
        final String prefix = PropertyManager.getProjectViewResourceBundle()
                .getString("uml.control_loop_diagram.label");
        diagram.setName(Facets.generateName(resourceSet, STAMPPlugin.FACET_SYMBOLIC_NAME,
                diagram.eClass(), prefix));
    }

    @Override
    protected String getNameString() {
        throw new UnsupportedOperationException();
    }

    @Override
    protected void postOpenEditor() {
        super.postOpenEditor();
        JomtGlobal.viewManager.getCurrentDiagramEditor().setSelected(true);
    }

    @Override
    protected void doPostProcessing() throws IllegalModelTypeException {
        final Diagram controlLoopDiagram = getDiagram();
        final TransactionalEditingDomain domain = TransactionUtil
                .getEditingDomain(controlLoopDiagram);
        final Diagram controlStructureDiagram = DiagramElementOperations.getDiagram(diagramElement);
        final LayoutCalculator layoutCalculator = new LayoutCalculator();
        // Control-Loop図上の位置は最後に調整するのでPaste位置は(0, 0)とする
        // 外部リンクやそれ付随する要素の追加は(0, 0)にPasteされたことを前提とするので、むやみに変更しないこと
        final Pnt2d pasteLocation = new Pnt2d(0, 0);
        final DiagramContext diagramContext = new DiagramContext(controlStructureDiagram,
                diagramSelections, toAbsolute(pasteLocation, diagramSelections, layoutCalculator));
        final CompoundCommand command = new StrictCompoundCommand();
        command.append(PasteCommand.create(domain, diagramContext, controlLoopDiagram,
                DIPackage.Literals.DIAGRAM_ELEMENT__OWNED_ELEMENT, copyResults,
                CommandParameter.NO_INDEX));

        // Copy時の座標からPaste位置へのVectorを算出
        final Bounds totalBounds = layoutCalculator.getTotalBounds(diagramSelections);
        final Vec2d pasteVector = new Vec2d(pasteLocation.getX() - totalBounds.getX(),
                pasteLocation.getY() - totalBounds.getY());

        for (final Edge edge : externalLinks) {
            final boolean souceIsExternal = isExternal(edge.getSource(), sourceElement,
                    targetElement);
            final DiagramElement externalEnd = souceIsExternal ? edge.getSource()
                    : edge.getTarget();
            final DiagramElement oppositeEnd = souceIsExternal ? edge.getTarget()
                    : edge.getSource();
            final DiagramElement copiedOppositeEnd = getCopiedElement(oppositeEnd);
            if (!(externalEnd instanceof Shape)) {
                continue;
            }

            final Shape copyEnd = appendCopyEndCommand(command, domain, (Shape) externalEnd,
                    oppositeEnd, edge, souceIsExternal, pasteVector, layoutCalculator);
            final Edge copyEdge = appendCopyEdgeCommand(command, domain, edge, copyEnd,
                    copiedOppositeEnd, souceIsExternal, pasteVector);
            appendCopySignalsCommand(command, domain, edge, copyEdge);
        }

        for (DiagramElement sel : Iterables.filter(diagramSelections, Shape.class)) {
            command.append(new DeferringCommandWrapper() {
                @Override
                protected Command createCommand() {
                    final Bounds selBounds = layoutCalculator.getBounds(sel);
                    final Bounds expectedBounds = DCFactory.eINSTANCE.createBounds(
                            selBounds.getX() + pasteVector.getX(),
                            selBounds.getY() + pasteVector.getY(), selBounds.getWidth(),
                            selBounds.getHeight());
                    final DiagramElement copiedElement = getCopiedElement(sel);
                    final Bounds copiedBounds = layoutCalculator.getBounds(copiedElement);
                    if (copiedElement != null && !DataTypes.eEquals(expectedBounds, copiedBounds)) {
                        return SetBoundsCommand.create(domain, null, copiedElement, expectedBounds);
                    }
                    return IdentityCommand.INSTANCE;
                }
            });
        }

        command.append(new DeferringCommandWrapper() {
            @Override
            protected Command createCommand() {
                // Control-Loop図上の図要素を見える位置に移動する
                final double marginX = 100;
                final double marginY = 100;
                final LayoutCalculator layoutCalculator = new LayoutCalculator();
                final List<DiagramElement> ownedElements = controlLoopDiagram.getOwnedElements();
                final Bounds totalBounds = layoutCalculator.getTotalBounds(ownedElements, true);
                final double x = marginX - totalBounds.getX();
                final double y = marginY - totalBounds.getY();
                final Vector vector = NotationFactory.eINSTANCE.createVector(x, y);
                return MoveTogetherCommand.create(domain, null, controlLoopDiagram, vector,
                        ownedElements);
            }
        });

        final IEntityStore doc = ModelGlobal.prjManager.getCurrentProject().doc;
        try {
            if (useTransaction) {
                doc.beginTransaction();
            }
            CommandUtils.execute(domain, command);
            if (useTransaction) {
                doc.commitTransaction();
            }
        } catch (final Exception ex) {
            LOGGER.error("error has occurred.", ex);
        }
    }

    private Pnt2d toAbsolute(final Pnt2d location, final Collection<DiagramElement> elements,
            final LayoutCalculator layoutCalculator) {
        final Pnt2d absolute = new Pnt2d(location);
        final Iterator<DiagramElement> iterator = elements.iterator();
        if (!iterator.hasNext()) {
            return absolute;
        }
        DiagramElement ancestor = iterator.next();
        while (iterator.hasNext()) {
            final DiagramElement element = iterator.next();
            ancestor = DiagramElements.getCommonAncestor(ancestor, element);
        }
        final Point boundsLocation = layoutCalculator.getBoundsLocation(ancestor);
        if (boundsLocation != null) {
            absolute.add(Geometries.toPnt2d(boundsLocation));
        }
        return absolute;
    }

    private Shape appendCopyEndCommand(final CompoundCommand command,
            final TransactionalEditingDomain domain, final Shape externalEnd,
            final DiagramElement oppositeEnd, final Edge edge, final boolean souceIsExternal,
            final Vec2d pasteVector, final LayoutCalculator layoutCalculator) {

        final Pnt2d location = decideLocation(externalEnd, oppositeEnd, edge, souceIsExternal,
                layoutCalculator);
        location.add(pasteVector);

        final Shape copyEnd = (Shape) STAMPNotationFactory.eINSTANCE.create(externalEnd.eClass());
        if (copyEnd instanceof STAMPComponentShape) {
            ((STAMPComponentShape) copyEnd).setIsOmitted(true);
        }
        ECollections.setEList(copyEnd.getModelElements(), externalEnd.getModelElements());

        final CreateShapeCommand createShapeCommand = new CreateShapeCommand();
        createShapeCommand.setOwningElement(decideOwningElement(externalEnd));
        createShapeCommand.setShape(copyEnd);
        createShapeCommand.setLocation(Geometries.toPoint(location));
        command.append(createShapeCommand.createTransactionalCommand(domain));

        return copyEnd;
    }

    @SuppressWarnings("deprecation")
    private Edge appendCopyEdgeCommand(final CompoundCommand command,
            final TransactionalEditingDomain domain, final Edge edge, final DiagramElement copyEnd,
            final DiagramElement copiedOppositeEnd, final boolean souceIsExternal,
            final Vec2d pasteVector) {
        final Edge copyEdge = STAMPNotationFactory.eINSTANCE.createSTAMPEdge();
        ECollections.setEList(copyEdge.getModelElements(), edge.getModelElements());

        final CreateEdgeCommand createEdgeCommand = new CreateEdgeCommand();
        final DiagramElement owningElement = decideOwningElement(edge);
        createEdgeCommand.setOwningElement(owningElement);
        createEdgeCommand.setEdge(copyEdge);
        if (souceIsExternal) {
            createEdgeCommand.setSource(copyEnd);
            createEdgeCommand.setTarget(copiedOppositeEnd);
        } else {
            createEdgeCommand.setSource(copiedOppositeEnd);
            createEdgeCommand.setTarget(copyEnd);
        }

        // 内部側のAnchorConstraintはコピーする、外部側はdefault
        if (!souceIsExternal && edge.getSourceAnchor() != null) {
            createEdgeCommand.setSourceAnchor(EcoreUtil.copy(edge.getSourceAnchor()));
        }
        if (souceIsExternal && edge.getTargetAnchor() != null) {
            createEdgeCommand.setTargetAnchor(EcoreUtil.copy(edge.getTargetAnchor()));
        }

        // LineConstraintを作成（制御点はコピーしないでおく）
        final LineConstraint copyLine = (LineConstraint) NotationFactory.eINSTANCE
                .create(edge.getLine().eClass());
        createEdgeCommand.setLine(copyLine);

        command.append(createEdgeCommand.createTransactionalCommand(domain));

        return copyEdge;
    }

    private DiagramElement decideOwningElement(final DiagramElement element) {
        return getDiagram();
    }

    private void appendCopySignalsCommand(final CompoundCommand command,
            final TransactionalEditingDomain domain, final Edge edge, final Edge copyEdge) {
        for (final STAMPLabel signal : Iterables.filter(
                Iterables.filter(edge.getOwnedElements(), STAMPLabel.class),
                label -> isSignal(label))) {
            final STAMPLabel copySignal = (STAMPLabel) STAMPNotationFactory.eINSTANCE
                    .create(signal.eClass());
            ECollections.setEList(copySignal.getModelElements(), signal.getModelElements());
            if (signal.getLocation() != null) {
                copySignal.setLocation(EcoreUtil.copy(signal.getLocation()));
            }

            final AddSignalCommand addSignalCommand = hasModel(signal, ControlAction.class)
                    ? new AddControlActionCommand()
                    : new AddFeedbackCommand();
            addSignalCommand.setParent((STAMPEdge) copyEdge);
            addSignalCommand.setShape(copySignal);
            command.append(addSignalCommand.createTransactionalCommand(domain));
            command.append(new SynchronizeLabelCommand(domain, copySignal));
        }
    }

    private DiagramElement getCopiedElement(final DiagramElement copySource) {
        for (final DiagramElement element : Iterables.filter(copyResults, DiagramElement.class)) {
            if (element.getModelElements().containsAll(copySource.getModelElements())) {
                return element;
            }
        }
        return null;
    }

    private Pnt2d decideLocation(final DiagramElement externalEnd, final DiagramElement oppositeEnd,
            final Edge edge, final boolean souceIsExternal,
            final LayoutCalculator layoutCalculator) {
        final Rectangle2d endBounds = Geometries
                .toRectangle2d(layoutCalculator.getBounds(externalEnd));
        // waypoints-head:source側 waypoints-tail:target側
        final List<Point> waypoints = layoutCalculator.getWaypoints(edge,
                layoutCalculator.getOrigin(edge));
        final Point firstPoint, secondPoint;
        if (souceIsExternal) {
            firstPoint = waypoints.get(waypoints.size() - 1);
            secondPoint = waypoints.get(0);
        } else {
            firstPoint = waypoints.get(0);
            secondPoint = waypoints.get(waypoints.size() - 1);
        }

        final Point center = DCFactory.eINSTANCE.createPoint(secondPoint);

        final double width = STAMPComponentShapeItemProvider.OMITTED_LAYOUT.getMinWidth();
        final double height = STAMPComponentShapeItemProvider.OMITTED_LAYOUT.getMinHeight();
        final double distanceX = 75 + width / 2;
        final double distanceY = 75 + height / 2;

        final int outcode = endBounds.outcode(center.getX(), center.getY());
        if (outcode != 0) {
            if ((outcode & Rectangle2D.OUT_LEFT) != 0) {
                center.setX(endBounds.getMinX() - distanceX);
            } else if ((outcode & Rectangle2D.OUT_RIGHT) != 0) {
                center.setX(endBounds.getMaxX() + distanceX);
            }

            if ((outcode & Rectangle2D.OUT_TOP) != 0) {
                center.setY(endBounds.getMinY() - distanceY);
            } else if ((outcode & Rectangle2D.OUT_BOTTOM) != 0) {
                center.setY(endBounds.getMaxY() + distanceY);
            }
        }
        final Pnt2d location = new Pnt2d(center.getX() - width / 2, center.getY() - height / 2);
        return location;
    }

    private void collectElements() {
        diagramSelections.clear();
        externalLinks.clear();
        sourceElement = null;
        targetElement = null;
        if (diagramElement instanceof STAMPEdge) {
            final STAMPDiagramElement source = ((STAMPEdge) diagramElement).getSTAMPSource();
            final STAMPDiagramElement target = ((STAMPEdge) diagramElement).getSTAMPTarget();
            sourceElement = source;
            targetElement = target;

            // 着目リンク、着目リンクの両端Element、両端Elementが同じ他リンクの収集（Copy&Paste対象）
            // （セルフリンクはCopyで自動的に含まれる）
            final Stream<DiagramElement> elementsByFocusedLink = Stream.concat(
                    Stream.concat(Stream.of(source), Stream.of(target)),
                    Stream.concat(
                            source.getSourceEdges().stream().filter(
                                    s -> s.getTarget() == target && s.getTarget() != s.getSource()),
                            source.getTargetEdges().stream().filter(
                                    s -> s.getSource() == target
                                            && s.getTarget() != s.getSource())));

            // 外部リンクで、始点・終点(LinkEnd)に接続するリンク（Copy&Paste対象）
            final Stream<DiagramElement> externalLinksToLinkEnd = Stream.concat(

                    Stream.concat(source.getSourceEdges().stream(),
                            target.getSourceEdges().stream())
                            .filter(e -> isLink(e)
                                    && isExternalLinkEnd(e.getTarget(), target, source)),

                    Stream.concat(source.getTargetEdges().stream(),
                            target.getTargetEdges().stream())
                            .filter(e -> isLink(e)
                                    && isExternalLinkEnd(e.getSource(), target, source)));

            // 外部リンクに接続する始点・終点(LinkEnd)（Copy&Paste対象）
            final Stream<DiagramElement> linkEndsByExternalLink = Stream.concat(

                    Stream.concat(source.getSourceEdges().stream(),
                            target.getSourceEdges().stream())
                            .filter(e -> isLink(e)
                                    && isExternalLinkEnd(e.getTarget(), target, source))
                            .map(e -> e.getTarget()),

                    Stream.concat(source.getTargetEdges().stream(),
                            target.getTargetEdges().stream())
                            .filter(e -> isLink(e)
                                    && isExternalLinkEnd(e.getSource(), target, source))
                            .map(e -> e.getSource()));

            // Copy&Paste対象群
            diagramSelections.addAll(Stream
                    .concat(elementsByFocusedLink,
                            Stream.concat(externalLinksToLinkEnd, linkEndsByExternalLink))
                    .distinct().sorted(EComparators.eObjectPath()).collect(toList()));

            // 外部リンクで、コンポーネントに接続するものの収集
            externalLinks.addAll(Stream
                    .concat(Stream
                            .concat(source.getSTAMPSourceEdges().stream(),
                                    target.getSTAMPSourceEdges().stream())
                            .filter(edge -> isLink(edge)
                                    && isExternalComponent(edge.getTarget(), target, source)),
                            Stream.concat(source.getSTAMPTargetEdges().stream(),
                                    target.getSTAMPTargetEdges().stream())
                                    .filter(edge -> isLink(edge) && isExternalComponent(
                                            edge.getSource(), target, source)))
                    .distinct()
                    .sorted(EComparators.eObjectPath())
                    .collect(toList()));
        }
    }

    private boolean hasModel(final DiagramElement element, final Class<?> type) {
        return DiagramElements.isModelElementPresent(element, type);
    }

    private boolean isLink(final DiagramElement element) {
        return hasModel(element, Link.class);
    }

    private boolean isLinkEnd(final DiagramElement element) {
        return hasModel(element, LinkEnd.class);
    }

    private boolean isSignal(final DiagramElement element) {
        return hasModel(element, Signal.class);
    }

    private boolean isComponent(final DiagramElement element) {
        return hasModel(element, Component.class);
    }

    private boolean isExternal(final DiagramElement end, final DiagramElement internalSource,
            final DiagramElement internalTarget) {
        return end != internalSource && end != internalTarget;
    }

    private boolean isExternalLinkEnd(final DiagramElement end, final DiagramElement internalSource,
            final DiagramElement internalTarget) {
        return isLinkEnd(end) && isExternal(end, internalSource, internalTarget);
    }

    private boolean isExternalComponent(final DiagramElement end,
            final DiagramElement internalSource, final DiagramElement internalTarget) {
        return isComponent(end) && isExternal(end, internalSource, internalTarget);
    }

    private boolean copyRelatedDiagramElements() {
        copyResults = null;
        if (!diagramSelections.isEmpty()) {
            final EditingDomain domain = ModelGlobal.prjManager.getCurrentProject().doc.getDomain();
            final Command copyCommand = CopyCommand.create(domain, diagramSelections);
            if (copyCommand.canExecute()) {
                copyCommand.execute();
                copyResults = copyCommand.getResult().stream()
                        .filter(DiagramElement.class::isInstance)
                        .collect(Collectors.toCollection(LinkedHashSet::new));
            }
        }
        return copyResults != null && !copyResults.isEmpty();
    }

    private ControlLoop getControlLoop(final HCFAnalysis hcfAnalysis) {
        final EObject element = DiagramElements.getModelElement(diagramElement);
        if (element != null && element instanceof ControlLink) {
            final Optional<ControlLoop> optional = hcfAnalysis.getControlLoops().stream()
                    .filter(loop -> loop.getTargetLink() == element).findFirst();
            if (optional.isPresent()) {
                return optional.get();
            } else {
                return hcfAnalysis.createControlLoop();
            }
        }
        return null;
    }

}
