package net.astah.stpa.stamp.ui.control;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.edit.domain.EditingDomain;

import com.google.common.collect.Iterables;

import net.astah.dd.di.Diagram;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.edit.command.Commands;
import net.astah.golf.model.IEntityStore;
import net.astah.jomt.jsystem.PropertyManager;
import net.astah.project.Facet;
import net.astah.project.Facets;
import net.astah.project.Project;
import net.astah.project.ProjectManagerAdapter;
import net.astah.stpa.stamp.ControlStructure;
import net.astah.stpa.stamp.STAMPPlugin;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.notation.ComponentExtractingTable;
import net.astah.stpa.stamp.notation.STAMPControlStructureDiagram;
import net.astah.stpa.stamp.notation.STAMPDiagram;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;
import net.astah.stpa.stamp.notation.edit.command.CreateSTAMPNotationRootCommand;
import net.astah.uml.exception.IllegalModelTypeException;

/**
 * コントロールストラクチャー図を作成するコマンドです。
 * <P>
 * 主に、マネジメントビューなどのメニューからのアクションイベントによる呼び出しを想定していてます。
 * </P>
 */
public class CreateControlStructureDiagramFromPrjCommand extends CreateSTAMPDiagramFromPrjCommand {

    @Override
    protected STAMPDiagram createNewDiagram(IEntityStore doc, EObject namedElement) {
        final EditingDomain domain = doc.getDomain();
        final Project project = ProjectManagerAdapter.getRootProject(namedElement);
        final Facet facet = project.getFacet(STAMPPlugin.FACET_SYMBOLIC_NAME);
        final STPAAnalysis analysis = (STPAAnalysis) facet.getContents().get(0);

        ControlStructure controlStructure = analysis.getControlStructures().stream().findFirst()
                .orElseGet(analysis::createControlStructure);

        STAMPControlStructureDiagram diagram = STAMPNotationFactory.eINSTANCE
                .createSTAMPControlStructureDiagram();
        Commands.execute(new CreateSTAMPNotationRootCommand(domain, facet, namedElement, diagram));
        DiagramElements.setModelElement(diagram, controlStructure);

        ComponentExtractingTable table = Iterables.getFirst(
                Iterables.filter(facet.getContents(), ComponentExtractingTable.class), null);
        if (table != null && table.getControlStructureDiagram() == null) {
            diagram.setComponentExtractingTable(table);
        }

        return diagram;
    }

    @Override
    protected void setDiagramName(IEntityStore doc, Diagram diagram) {
        ResourceSet resourceSet = doc.getDomain().getResourceSet();
        String prefix = PropertyManager.getProjectViewResourceBundle()
                .getString("uml.control_structure_diagram.label");
        diagram.setName(Facets.generateName(resourceSet, STAMPPlugin.FACET_SYMBOLIC_NAME,
                diagram.eClass(), prefix));
    }

    @Override
    protected String getNameString() {
        throw new UnsupportedOperationException();
    }

    @Override
    protected void doPostProcessing() throws IllegalModelTypeException {
        // nothing
    }

}
