package net.astah.stpa.stamp.ui.control;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.emf.edit.job.CommandJob;
import net.astah.emf.edit.job.Jobs;
import net.astah.golf.draw.GNode;
import net.astah.golf.widget.CCanvas;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.DiagramEditor;
import net.astah.jomt.jview.DrawSuggest;
import net.astah.notation.editor.IViewElement;
import net.astah.notation.editor.ViewElements;
import net.astah.stpa.stamp.notation.STAMPCompartment;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;
import net.astah.stpa.stamp.notation.STAMPNotationPackage;

/**
 * コンパートメントの要素を追加するコマンドです。
 */
public class CreateElementInCompartmentCommand extends net.astah.golf.control.Command {

    private STAMPCompartment compartment;

    public CreateElementInCompartmentCommand() {
    }

    public void setCompartment(STAMPCompartment compartment) {
        this.compartment = compartment;
    }

    @Override
    public void execute() {
        if (compartment == null) {
            compartment = getSelectedCompartment();
            if (compartment == null) {
                return;
            }
        }

        TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(compartment);

        // このコマンドが実行するEMFコマンドを構築する
        Command command = createTransactionalCommand(domain);

        // 構築したEMFコマンドを非同期ジョブとして実行する
        CommandJob job = new CommandJob(domain, command);
        Jobs.toCompletableFuture(job, true).thenRun(this::postExecute);
        job.schedule();
    }

    /**
     * トランザクションとして実行するモデル編集コマンドの作成.
     * 
     * @param domain
     *            コマンドを実行するドメイン
     * @return 作成したコマンド
     */
    public Command createTransactionalCommand(TransactionalEditingDomain domain) {
        STAMPLabel label = STAMPNotationFactory.eINSTANCE.createSTAMPLabel();
        label.setText(String.format("Element%d", compartment.getElementInCompartments().size()));
        return AddCommand.create(domain, compartment,
                STAMPNotationPackage.Literals.STAMP_COMPARTMENT__ELEMENT_IN_COMPARTMENT, label);
    }

    private STAMPCompartment getSelectedCompartment() {
        Object source = event.getSource();
        if (source instanceof DrawSuggest) {
            Notifier model = ((DrawSuggest) source).getOwner().getModel();
            if (model instanceof STAMPCompartment) {
                return (STAMPCompartment) model;
            }
        }
        return null;
    }

    /**
     * トランザクション完了後の後処理.
     */
    protected void postExecute() {
        DiagramEditor editor = JomtGlobal.viewManager.getCurrentDiagramEditor();
        CCanvas canvas = editor.getCanvas();
        canvas.unselect();

        for (IViewElement iViewElement : ViewElements.viewElementsOf(compartment)) {
            canvas.selectAndNotify(((GNode) iViewElement).getChildren());
            canvas.repaint();
        }
    }

}
