package net.astah.stpa.stamp.ui.control;

import static net.astah.stpa.stamp.notation.STAMPNotationPackage.Literals.UCA_TABLE__CONTROL_STRUCTURE_DIAGRAM;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.edit.domain.EditingDomain;

import com.google.common.base.Strings;
import com.google.common.collect.Iterables;
import com.google.common.collect.Iterators;
import com.google.common.collect.Lists;

import net.astah.common.collect.Streams;
import net.astah.dd.dc.Point;
import net.astah.dd.di.Diagram;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.Edge;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.common.ecore.EComparators;
import net.astah.emf.common.notify.EFeaturePath;
import net.astah.emf.edit.command.Commands;
import net.astah.golf.control.Command;
import net.astah.golf.model.IEntityStore;
import net.astah.golf.view.UIView;
import net.astah.jomt.jcontrol.OpenTableEditorCommand;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jsystem.PropertyManager;
import net.astah.jomt.jview.DiagramEditor;
import net.astah.jomt.jview.MessageManager;
import net.astah.jomt.jview.StructureView;
import net.astah.project.Facet;
import net.astah.project.Facets;
import net.astah.project.Project;
import net.astah.project.ProjectManagerAdapter;
import net.astah.stpa.stamp.ControlAction;
import net.astah.stpa.stamp.ControlLink;
import net.astah.stpa.stamp.STAMPFactory;
import net.astah.stpa.stamp.STAMPPlugin;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.UCAAnalysis;
import net.astah.stpa.stamp.notation.STAMPCompartmentableShape;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.notation.STAMPControlStructureDiagram;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;
import net.astah.stpa.stamp.notation.UCATable;
import net.astah.stpa.stamp.notation.edit.command.CreateSTAMPNotationRootCommand;
import net.astah.stpa.stamp.ui.Messages;
import net.astah.uml.exception.IllegalModelTypeException;

/**
 * UCA 表を作成するコマンドです。
 */
public class CreateUCATableFromPrjCommand extends CreateSTAMPDiagramFromPrjCommand {

    @Override
    protected Diagram createNewDiagram(final IEntityStore doc, final EObject owner) {
        final STAMPControlStructureDiagram selectedCSDiagram = getSelectedControlStructureDiagramInStructureView();

        final EditingDomain domain = doc.getDomain();
        final Project project = ProjectManagerAdapter.getRootProject(owner);
        final Facet facet = project.getFacet(STAMPPlugin.FACET_SYMBOLIC_NAME);
        final STPAAnalysis analysis = (STPAAnalysis) facet.getContents().get(0);

        STAMPControlStructureDiagram controlStructureDiagram = null;
        if (selectedCSDiagram != null) {
            controlStructureDiagram = selectedCSDiagram;
        } else {
            final List<STAMPControlStructureDiagram> diagrams = Streams
                    .stream(facet.getContents(), STAMPControlStructureDiagram.class)
                    .collect(Collectors.toList());
            if (diagrams.size() == 0) {
                MessageManager.displayConfirmMessageDialog("",
                        Messages.create_uca_table_confirm_no_cs);
            } else if (diagrams.size() == 1) {
                controlStructureDiagram = diagrams.get(0);
            } else {
                final DiagramEditor currentEditor = JomtGlobal.viewManager
                        .getCurrentDiagramEditor();
                if (currentEditor != null
                        && currentEditor.getDiagram() instanceof STAMPControlStructureDiagram) {
                    controlStructureDiagram = (STAMPControlStructureDiagram) JomtGlobal.viewManager
                            .getCurrentDiagramEditor().getDiagram();
                } else {
                    MessageManager.displayConfirmMessageDialog("",
                            isStructureViewSelected() ? Messages.create_uca_table_confirm_many_cs
                                    : Messages.create_uca_table_confirm_many_cs_2);
                }
            }
        }

        if (controlStructureDiagram == null) {
            return null;
        }

        // UCA表が作成済みの場合には、すでにあるUCA表を開く
        final UCATable ucaTable = getUCATable(controlStructureDiagram);
        if (ucaTable != null) {
            return ucaTable;
        }

        UCAAnalysis ucaAnalysis = analysis.getUcaAnalysis();
        if (ucaAnalysis == null) {
            ucaAnalysis = STAMPFactory.eINSTANCE.createUCAAnalysis();
        }
        analysis.setUcaAnalysis(ucaAnalysis);
        ucaAnalysis.setGuideWordSet(analysis.getUCAGuideWordSets().get(0));

        // ControlActionの順序を記憶する
        registerControlActionsOrder(controlStructureDiagram, ucaAnalysis);

        final UCATable table = STAMPNotationFactory.eINSTANCE.createUCATable();
        Commands.execute(new CreateSTAMPNotationRootCommand(domain, facet, owner, table));
        table.setAnalysis(ucaAnalysis);
        table.setControlStructureDiagram(controlStructureDiagram);

        return table;
    }

    private STAMPControlStructureDiagram getSelectedControlStructureDiagramInStructureView() {
        if (!isStructureViewSelected()) {
            return null;
        }

        final Set<Object> selectedModelElements = JomtGlobal.viewManager.getProjectView()
                .getSelectedModelElements();
        if (selectedModelElements == null) {
            return null;
        }

        final List<STAMPControlStructureDiagram> csDiagrams = Lists.newArrayList(
                Iterables.filter(selectedModelElements, STAMPControlStructureDiagram.class));
        if (csDiagrams.isEmpty()) {
            return null;
        } else if (csDiagrams.size() > 1) {
            csDiagrams.sort(EComparators.eObjectPath());
        }

        return csDiagrams.get(0);
    }

    private boolean isStructureViewSelected() {
        final UIView selectedUIViewTab = JomtGlobal.viewManager.getProjectView()
                .getSelectedUIViewTab();
        return selectedUIViewTab instanceof StructureView;
    }

    private UCATable getUCATable(final STAMPControlStructureDiagram controlStructureDiagram) {
        final EFeaturePath path = EFeaturePath.build(UCA_TABLE__CONTROL_STRUCTURE_DIAGRAM);
        return path.getInverseReferences(controlStructureDiagram).stream()
                .map(UCATable.class::cast)
                .findFirst().orElse(null);
    }

    private void registerControlActionsOrder(final STAMPControlStructureDiagram diagram,
            final UCAAnalysis analysis) {
        // 図上のコンポーネントを取得し、ソートする(x, y座標)
        final List<STAMPCompartmentableShape> componentShapes = Lists.newArrayList(Iterators
                .filter(DiagramElements.getAllContents(diagram), STAMPComponentShape.class));
        final Comparator<DiagramElement> locationComparator = EComparators.cast(
                DiagramElement.class,
                new ElementLocationComparator());
        Collections.sort(componentShapes, locationComparator);
        // コンポーネントをsourceに持つEdgeを作成順にソートする
        for (final STAMPCompartmentableShape shape : componentShapes) {
            final EList<Edge> sourceEdges = shape.getSourceEdges();
            final ArrayList<Edge> sortedEdges = Lists.newArrayList(sourceEdges);
            sortedEdges.sort((o1, o2) -> {
                // EComparators.cast()しているため、NPEは発生しない
                final EObject m1 = DiagramElements.getModelElement(o1);
                final EObject m2 = DiagramElements.getModelElement(o2);
                return EComparators.eObjectPath().compare(m1, m2);
            });
            final List<ControlAction> actions = sortedEdges.stream()
                    .map(edge -> DiagramElements.getModelElement(edge, ControlLink.class))
                    .filter(Objects::nonNull).flatMap(link -> link.getActions().stream())
                    .collect(Collectors.toList());
            analysis.getControlActions().addAll(actions);
        }
    }

    @Override
    protected void setDiagramName(final IEntityStore doc, final Diagram diagram) {
        final ResourceSet resourceSet = doc.getDomain().getResourceSet();
        if (!Strings.isNullOrEmpty(diagram.getName())) {
            return;
        }
        final String prefix = PropertyManager.getProjectViewResourceBundle()
                .getString("uml.uca_table.label");
        diagram.setName(Facets.generateName(resourceSet, STAMPPlugin.FACET_SYMBOLIC_NAME,
                diagram.eClass(), prefix));
    }

    @Override
    protected String getNameString() {
        throw new UnsupportedOperationException();
    }

    @Override
    protected void doPostProcessing() throws IllegalModelTypeException {
        // 親の処理を無効化
    }

    @Override
    protected Command createOpenEditorCommand(final Diagram diagram) {
        final OpenTableEditorCommand command = new OpenTableEditorCommand();
        command.setModel(diagram);
        return command;
    }

    static class ElementLocationComparator implements Comparator<DiagramElement> {

        private final LayoutCalculator layoutCalculator = new LayoutCalculator();

        @Override
        public int compare(final DiagramElement o1, final DiagramElement o2) {
            return compareLocation(o1, o2);
        }

        private int compareLocation(final DiagramElement shape1, final DiagramElement shape2) {
            final Point point1 = layoutCalculator.getBoundsLocation(shape1);
            final Point point2 = layoutCalculator.getBoundsLocation(shape2);
            final double distanceX = point1.getX() - point2.getX();
            if (distanceX != 0) {
                return distanceX > 0 ? 1 : -1;
            }
            final double distanceY = point1.getY() - point2.getY();
            if (distanceY != 0) {
                return distanceY > 0 ? 1 : -1;
            }
            return 0;
        }
    }
}
