package net.astah.stpa.stamp.ui.control;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.transaction.TransactionalEditingDomain;

import net.astah.golf.model.IEntityStore;
import net.astah.jomt.jsystem.ModelGlobal;
import net.astah.stpa.stamp.ControlStructure;

/**
 * コンポーネントの分析対象を参照し、処理するコマンドです。
 */
public class ModifyAnalysisTargetCommand extends net.astah.golf.control.Command {

    private ControlStructure controlStructure;

    @Override
    public void execute() {
        if (ModelGlobal.prjManager == null || ModelGlobal.prjManager.getCurrentProject() == null) {
            return;
        }
        IEntityStore doc = ModelGlobal.prjManager.getCurrentProject().doc;
        if (doc == null) {
            return;
        }
        if (controlStructure == null) {
            return;
        }

        TransactionalEditingDomain domain = (TransactionalEditingDomain) doc.getDomain();
        boolean isInTransaction = doc.isInTransaction();
        if (isInTransaction) {
            return;
        }

        try {
            doc.beginTransaction();

            Command command = new ModifyControlStructureDiagramElementsCommand(domain,
                    controlStructure);
            if (command.canExecute()) {
                command.execute();
            }

            doc.commitTransaction();
        } catch (Exception e) {
            throw e;
        } finally {
            doc.abortTransaction();
        }
    }

    /**
     * 処理するCS図をセットします。
     * 
     * @param controlStructure
     *            CS図
     */
    public void setControlStructure(ControlStructure controlStructure) {
        this.controlStructure = controlStructure;
    }

}
