package net.astah.stpa.stamp.ui.menu.handler;

import java.awt.event.ActionEvent;
import java.util.List;

import org.eclipse.emf.ecore.EObject;

import net.astah.dd.di.util.DiagramElements;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.swing.Selections;
import net.astah.stpa.stamp.ProcessVariable;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.ui.control.AddProcessValueCommand;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * 図上でプロセスモデル変数値を追加します。
 */
@Handler(targets = { STAMPLabel.class })
@MenuItem(label = "%StampControlStructureDiagram_addProcessValue_label",
        tooltip = "%StampControlStructureDiagram_addProcessValue_tooltip",
        type = ItemKind.PUSH)
@MenuCommand(id = "stpa.AddProcessValueCommand",
        category = "edit.add_process_model")
public class AddProcessValueHandler {
    /**
     * プロセスモデル変数値を追加するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        final STAMPLabel label = (STAMPLabel) Selections.getFirstElement(selection);
        final AddProcessValueCommand command = new AddProcessValueCommand();
        command.setOwnerDiagramElement(label);
        command.execute();
    }

    /**
     * プロセスモデル変数値を追加するコマンドが実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        if (Selections.size(selection) != 1) {
            return false;
        }
        final STAMPLabel label = (STAMPLabel) Selections.getFirstElement(selection);
        final EObject model = DiagramElements.getModelElement(label);
        return model instanceof ProcessVariable;
    }
}
