package net.astah.stpa.stamp.ui.menu.handler;

import java.awt.event.ActionEvent;
import java.util.List;

import net.astah.dd.dg.Ellipse;
import net.astah.dd.dg.Rectangle;
import net.astah.dd.di.util.DiagramElements;
import net.astah.golf.control.GolfAction;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.swing.Selections;
import net.astah.stpa.stamp.LinkEnd;
import net.astah.stpa.stamp.Note;
import net.astah.stpa.stamp.notation.STAMPCompartmentableShape;
import net.astah.stpa.stamp.notation.STAMPShape;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * 選択した図要素の背景色を設定します。
 */
@Handler
@MenuItem(label = "%StampControlStructureDiagram_setColor_label",
        iconURI = "platform:/plugin/net.astah.core.app/net/astah/jomt/resource/images/setcolor.png",
        tooltip = "%StampControlStructureDiagram_setColor_tooltip",
        type = ItemKind.CHECK)
@MenuCommand(id = "stpa.SetColorCommand",
        category = "edit.color")
public class SetColorHandler {

    /**
     * 選択した図要素の背景色を設定するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        final GolfAction action = new GolfAction("SetColor");
        action.actionPerformed(e);
    }

    /**
     * 選択した図要素の背景色を設定するコマンドが実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        // ターゲット要素が1つでもあれば、true
        return Selections.stream(selection).anyMatch(object -> isTarget(object));
    }

    private boolean isTarget(final Object object) {
        return object instanceof STAMPCompartmentableShape // コンポーネントなど
                || isLinkEnd(object)
                || object instanceof Rectangle // 長方形、テキスト
                || object instanceof Ellipse // 楕円
                || isNote(object); // ノート
    }

    private boolean isLinkEnd(final Object object) {
        return object instanceof STAMPShape
                && DiagramElements.isModelElementPresent((STAMPShape) object, LinkEnd.class);
    }

    private boolean isNote(final Object object) {
        return object instanceof STAMPShape
                && DiagramElements.isModelElementPresent((STAMPShape) object, Note.class);
    }

}
