package net.astah.stpa.stamp.ui.table;

import static net.astah.stpa.stamp.STAMPPackage.Literals.COMPONENT;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

import org.eclipse.emf.common.notify.AdapterFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.edit.provider.IItemPropertyDescriptor;
import org.eclipse.emf.edit.provider.ItemPropertyDescriptor;

import net.astah.emf.common.ecore.EObjects;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.ControlStructure;
import net.astah.stpa.stamp.Feedback;
import net.astah.stpa.stamp.FeedbackLink;
import net.astah.stpa.stamp.LinkableElement;
import net.astah.table.CellDescriptor;
import net.astah.table.ICellEvaluator;
import net.astah.table.Tuple;

/**
 * コンポーネント抽出表のフィードバック列を表示するための{@link IItemPropertyDescriptor} を定義したクラスです。
 */
public class FeedbackCellEvaluator implements ICellEvaluator {

    @Override
    public Object getElement(final CellDescriptor cellDescriptor, final Tuple tuple) {
        return null;
    }

    @Override
    public IItemPropertyDescriptor getHeaderPropertyDescriptor(final CellDescriptor cellDescriptor,
            final Tuple tuple) {
        return null;
    }

    @Override
    public IItemPropertyDescriptor getCellPropertyDescriptor(final CellDescriptor cellDescriptor,
            final Tuple rowTuple, final Tuple columnTuple) {
        final EObject eObject = rowTuple.getElements().get(0).getObject();
        if (!(eObject instanceof Component)) {
            return null;
        }
        final Component component = (Component) eObject;
        final AdapterFactory adapterFactory = ItemProviders.getAdapterFactory(component);
        return new FeedbackPropertyDescriptor(adapterFactory, "", "", null, component);
    }

    @Override
    public Comparator<?> getComparator(final CellDescriptor cellDescriptor) {
        return null;
    }

    /**
     * フィードバックのプロパティ取得用{@link ItemPropertyDescriptor} です。
     */
    class FeedbackPropertyDescriptor extends ItemPropertyDescriptor {

        private final Component component;

        public FeedbackPropertyDescriptor(final AdapterFactory adapterFactory,
                final String displayName, final String description,
                final EReference[] parentReferences, final Component component) {
            super(adapterFactory, displayName, description, parentReferences);
            this.component = component;

        }

        @Override
        public Object getPropertyValue(final Object object) {
            List<FeedbackLink> list = EObjects.getAncestor(ControlStructure.class, component)
                    .getLinks().stream()
                    .filter(m -> m instanceof FeedbackLink
                            && ((FeedbackLink) m).getSource().equals(component))
                    .map(m -> (FeedbackLink) m).collect(Collectors.toList());

            List<Feedback> actions = new ArrayList<>();
            for (FeedbackLink link : list) {
                final LinkableElement source = link.getSource();
                final LinkableElement target = link.getTarget();
                if (!COMPONENT.isInstance(source) || !COMPONENT.isInstance(target)) {
                    continue;
                }
                actions.addAll(link.getFeedbacks());
            }

            return actions;
        }
    }
}