package net.astah.stpa.stamp.ui.table;

import java.util.Collection;
import java.util.Comparator;

import org.eclipse.emf.common.notify.AdapterFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.eclipse.emf.edit.provider.IItemPropertyDescriptor;
import org.eclipse.emf.edit.provider.ItemPropertyDescriptor;
import org.eclipse.uml2.common.util.UML2Util;

import com.google.common.base.Optional;
import com.google.common.collect.FluentIterable;

import net.astah.emf.edit.provider.ItemProviders;
import net.astah.stpa.stamp.ControlAction;
import net.astah.stpa.stamp.UCAAnalysisResult;
import net.astah.stpa.stamp.UCAGuideWord;
import net.astah.table.CellDescriptor;
import net.astah.table.ICellEvaluator;
import net.astah.table.Tuple;

/**
 * UCA表のUCA列を表示するための{@link IItemPropertyDescriptor} を定義したクラスです。
 */
public class UCAAnalysisResultCellEvaluator implements ICellEvaluator {

    @Override
    public Object getElement(final CellDescriptor cellDescriptor, final Tuple tuple) {
        return null;
    }

    @Override
    public IItemPropertyDescriptor getHeaderPropertyDescriptor(final CellDescriptor cellDescriptor,
            final Tuple tuple) {
        return null;
    }

    @Override
    public IItemPropertyDescriptor getCellPropertyDescriptor(final CellDescriptor cellDescriptor,
            final Tuple rowTuple, final Tuple columnTuple) {
        final EObject columnObject = columnTuple.getElements().get(0).getObject();
        if (!(columnObject instanceof UCAGuideWord)) {
            return null;
        }
        final UCAGuideWord guideWord = (UCAGuideWord) columnObject;

        final EObject rowObject = rowTuple.getElements().get(0).getObject();
        if (!(rowObject instanceof ControlAction)) {
            return null;
        }
        final ControlAction action = (ControlAction) rowObject;
        final Collection<Setting> inverseReferences = UML2Util.getInverseReferences(action);
        final Optional<UCAAnalysisResult> ucaAnalysisResultOption = FluentIterable
                .from(inverseReferences)
                .transform(Setting::getEObject)
                .filter(UCAAnalysisResult.class)
                .filter(result -> result.getGuideWord() == guideWord)
                .first();

        final AdapterFactory adapterFactory = ItemProviders.getAdapterFactory(action);
        final String displayName = "";
        final String description = "";
        final EReference[] parentReference = null;
        return new UCAPropertyDescriptor(adapterFactory, displayName, description,
                parentReference, ucaAnalysisResultOption.orNull());
    }

    @Override
    public Comparator<?> getComparator(final CellDescriptor cellDescriptor) {
        return null;
    }

    /**
     * UCA表のUCAのプロパティ取得用{@link ItemPropertyDescriptor} です。
     */
    class UCAPropertyDescriptor extends ItemPropertyDescriptor {
        private Object ucaResultOrAnalysis = null;

        public UCAPropertyDescriptor(final AdapterFactory adapterFactory, final String displayName,
                final String description, final EReference[] parentReferences,
                final Object ucaResultOrAnalysis) {
            super(adapterFactory, displayName, description, parentReferences);
            this.ucaResultOrAnalysis = ucaResultOrAnalysis;
        }

        @Override
        public Object getPropertyValue(final Object object) {
            if (ucaResultOrAnalysis != null) {
                return ((UCAAnalysisResult) ucaResultOrAnalysis).getResultElements();
            }
            return null;
        }
    }
}
