package net.astah.stpa.stamp.ui.view;

import java.awt.Color;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Collections;
import java.util.Enumeration;

import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.ToolTipManager;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

import org.eclipse.emf.ecore.EObject;

import net.astah.emf.edit.provider.swing.JTreeBuilder;
import net.astah.golf.control.CommandEvent;
import net.astah.golf.control.CommandManager;
import net.astah.golf.draw.GCamera;
import net.astah.golf.model.EntityStoreEvent;
import net.astah.golf.model.EntityStoreListener;
import net.astah.golf.model.IEntityStore;
import net.astah.golf.util.ResourceBundleList;
import net.astah.golf.util.UIFactory;
import net.astah.golf.view.GroupUIView;
import net.astah.golf.view.ModelUIView;
import net.astah.golf.widget.swing.SwingComponent;
import net.astah.jmodel.DiagramViewInfo;
import net.astah.jomt.jcontrol.OpenDiagramEditorCommand;
import net.astah.jomt.jcontrol.OpenTableEditorCommand;
import net.astah.jomt.jsystem.ModelGlobal;
import net.astah.notation.Diagram;
import net.astah.notation.TableView;
import net.astah.project.Facets;
import net.astah.project.Project;
import net.astah.project.ProjectManagerAdapter;
import net.astah.stpa.stamp.STAMPPlugin;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.UCAAnalysis;
import net.astah.stpa.stamp.notation.AccidentHazardSafetyConstraintTable;
import net.astah.stpa.stamp.notation.ComponentExtractingTable;
import net.astah.stpa.stamp.notation.CountermeasureTable;
import net.astah.stpa.stamp.notation.HCFTable;
import net.astah.stpa.stamp.notation.PreconditionTable;
import net.astah.stpa.stamp.notation.STAMPControlLoopDiagram;
import net.astah.stpa.stamp.notation.STAMPControlStructureDiagram;
import net.astah.stpa.stamp.notation.UCATable;
import net.astah.stpa.stamp.notation.util.STAMPNotationSwitch;
import net.astah.stpa.stamp.ui.Messages;

/**
 * STAMP分析手順ツリービュークラスです。
 */
public class AnalysisProcedureView extends ModelUIView
        implements MouseListener, EntityStoreListener {
    private final JTree tree;

    private static final Color PREPARED_COLOR = Color.BLUE;

    private static final Color NON_PREPARED_COLOR = Color.BLACK;

    private static final Color SELECTED_COLOR = Color.WHITE;

    private final int defaultDismissTimeout = ToolTipManager.sharedInstance().getDismissDelay();

    String packageNode = "platform:/plugin/net.astah.stpa.stamp.notation.edit/icons/full/obj16/PackageNode.png";
    String leafNode = "platform:/plugin/net.astah.stpa.stamp.notation.edit/icons/full/obj16/LeafNode.png";

    /**
     * 分析手順ビューを構築します。
     * 
     * @param doc
     *            表示するモデルの {@code IEntityStore}
     */
    public AnalysisProcedureView(IEntityStore doc) {
        UIFactory uiFactory = new UIFactory(new ResourceBundleList());
        doc.addEntityStoreListener(this);
        tree = new JTree() {
            private static final long serialVersionUID = 1L;

            @Override
            public String getToolTipText(MouseEvent event) {
                TreePath path = getPathForLocation(event.getX(), event.getY());
                if (path != null) {
                    DefaultMutableTreeNode node = (DefaultMutableTreeNode) path
                            .getLastPathComponent();
                    TreeUserObject treeUserObject = (TreeUserObject) node.getUserObject();
                    return treeUserObject.getToolTipText();
                }

                return null;
            }
        };

        tree.addMouseListener(new MouseAdapter() {

            @Override
            public void mouseEntered(final MouseEvent me) {
                ToolTipManager.sharedInstance().setDismissDelay(Integer.MAX_VALUE);
            }

            @Override
            public void mouseExited(final MouseEvent me) {
                ToolTipManager.sharedInstance().setDismissDelay(defaultDismissTimeout);
            }
        });

        tree.setCellRenderer(new DefaultTreeCellRenderer() {
            private static final long serialVersionUID = 1L;

            @Override
            public void updateUI() {
                super.updateUI();
                setOpenIcon(uiFactory.getResourceImageIcon(packageNode));
                setClosedIcon(uiFactory.getResourceImageIcon(packageNode));
                setLeafIcon(uiFactory.getResourceImageIcon(leafNode));
            }

            @Override
            public java.awt.Component getTreeCellRendererComponent(JTree tree, Object value,
                    boolean selected, boolean expanded, boolean leaf, int row, boolean hasFocus) {
                super.getTreeCellRendererComponent(tree, value, selected, expanded, leaf, row,
                        hasFocus);
                setForeground(getNodeForegroundColor(value, selected));
                return this;
            }
        });
        ToolTipManager.sharedInstance().registerComponent(tree);
        tree.setModel(getTreeModel());
        tree.setLargeModel(true);
        init();
    }

    private void init() {
        final JScrollPane scrollPane = new JScrollPane(tree);
        final SwingComponent baseComponent = new SwingComponent(scrollPane);
        setBaseComponent(baseComponent);
        component = baseComponent;
        addComponent(baseComponent);

        expandAll(tree, new TreePath(tree.getModel().getRoot()));
        tree.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                if (SwingUtilities.isLeftMouseButton(e) && e.getClickCount() == 2) {
                    JTree tree = (JTree) e.getSource();
                    TreePath path = tree.getPathForLocation(e.getX(), e.getY());
                    openOrCreateDiagram(path);
                }
            }
        });
    }

    /**
     * 指定したツリーのノード以下をすべて展開状態にします。
     * 
     * @param tree
     *            対象ツリー
     * @param parent
     *            対象ノードパス
     */
    private void expandAll(JTree tree, TreePath parent) {
        TreeNode node = (TreeNode) parent.getLastPathComponent();
        Collections.list((Enumeration<?>) node.children()).stream()
                .filter(TreeNode.class::isInstance).map(TreeNode.class::cast)
                .forEach(n -> expandAll(tree, parent.pathByAddingChild(n)));
        tree.expandPath(parent);
    }

    @Override
    public String getViewTitle() {
        return Messages.AnalysisProcedureView_title;
    }

    @Override
    public void setParent(final GroupUIView p) {
        super.setParent(p);
        if (p == null) {
            JTreeBuilder.dispose(tree);
        }
    }

    /**
     * ツリー表示するモデルを取得します。
     * 
     * @return TreeModel Treeに設定するTreeModel
     */
    protected TreeModel getTreeModel() {
        final DiagramNodePredicate predicate = new DiagramNodePredicate();

        DefaultMutableTreeNode root = new DefaultMutableTreeNode(new TreeUserObject(
                Messages.AnalysisProcedureView_stpa_analysis_procedure_title, predicate));

        DefaultMutableTreeNode step0 = new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_step_0_title, predicate));
        root.add(step0);
        DefaultMutableTreeNode preparation1 = new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_preparation_1_title, predicate));
        step0.add(preparation1);

        preparation1.add(new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_precondition_title,
                        Messages.AnalysisProcedureView_CreatePreconditionCommand,
                        PreconditionTable.class,
                        Messages.AnalysisProcedureView_precondition_tooltip, predicate)));
        preparation1.add(new DefaultMutableTreeNode(new TreeUserObject(
                Messages.AnalysisProcedureView_accident_hazard_safety_constraint_title,
                Messages.AnalysisProcedureView_CreateAccidentHazardSafetyConstraintCommand,
                AccidentHazardSafetyConstraintTable.class,
                Messages.AnalysisProcedureView_accident_hazard_safety_constraint_tooltip,
                predicate)));
        preparation1.add(new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_component_extracting_title,
                        Messages.AnalysisProcedureView_CreateComponentExtractingCommand,
                        ComponentExtractingTable.class,
                        Messages.AnalysisProcedureView_component_extracting_tooltip, predicate)));

        DefaultMutableTreeNode preparation2 = new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_preparation_2_title, predicate));
        step0.add(preparation2);
        preparation2.add(new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_control_structure_title,
                        Messages.AnalysisProcedureView_CreateControlStructureDiagramCommand,
                        STAMPControlStructureDiagram.class,
                        Messages.AnalysisProcedureView_control_structure_tooltip, predicate)));

        DefaultMutableTreeNode step1 = new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_step_1_title, predicate));
        root.add(step1);
        step1.add(new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_uca_title,
                        Messages.AnalysisProcedureView_CreateUCATableCommand, UCATable.class,
                        Messages.AnalysisProcedureView_uca_tooltip, predicate)));

        DefaultMutableTreeNode step2 = new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_step_2_title, predicate));
        root.add(step2);
        DefaultMutableTreeNode hcf = new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_hcf_title,
                        Messages.AnalysisProcedureView_hcf_tooltip));
        step2.add(hcf);
        hcf.add(new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_control_loop_diagram_title,
                        Messages.AnalysisProcedureView_CreateControlLoopDiagramCommand,
                        STAMPControlLoopDiagram.class,
                        Messages.AnalysisProcedureView_control_loop_diagram_tooltip, predicate)));
        hcf.add(new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_hcf_diagram_title,
                        Messages.AnalysisProcedureView_CreateHCFTableCommand, HCFTable.class,
                        Messages.AnalysisProcedureView_hcf_diagram_tooltip,
                        predicate)));

        DefaultMutableTreeNode step3 = new DefaultMutableTreeNode(
                new TreeUserObject(Messages.AnalysisProcedureView_countermeasure_title,
                        Messages.AnalysisProcedureView_CreateCountermeasureTableCommand,
                        CountermeasureTable.class,
                        Messages.AnalysisProcedureView_countermeasure_tooltip, predicate));
        root.add(step3);

        return new DefaultTreeModel(root);
    }

    /**
     * 対象のツリーパスのUserObjectからDiagramを開くまたは作成します。
     * 
     * @param path
     *            TreePath
     */
    private void openOrCreateDiagram(TreePath path) {
        if (path == null) {
            return;
        }
        DefaultMutableTreeNode node = (DefaultMutableTreeNode) path.getLastPathComponent();
        if (node.getUserObject() != null) {
            TreeUserObject treeUserObject = (TreeUserObject) node.getUserObject();
            if (!treeUserObject.isDiagramNode()) {
                return;
            }
            Diagram diagram = treeUserObject.getDiagram();
            if (diagram != null) {
                if (isCreateNewDiagram(diagram)) {
                    treeUserObject.createDiagram();
                } else if (diagram instanceof TableView) {
                    openMatrixDiagram(diagram);
                } else {
                    openDiagram(diagram);
                }
            } else {
                treeUserObject.createDiagram();
            }
        }
    }

    /**
     * 指定したClassで現在開いているプロジェクト内のDiagramを取得します。
     * 
     * @param diagramClass
     *            探索対象Diagram Class
     * @return プロジェクト内で最初に見つかったDiagram、見つからなかった場合{@code null}
     */
    private Diagram getDiagram(Class<? extends Diagram> diagramClass) {
        IEntityStore doc = ModelGlobal.prjManager.getCurrentProject().doc;
        Project project = ProjectManagerAdapter.getRootProject(doc.getDomain().getResourceSet());
        return Facets.getFirstContent(project, STAMPPlugin.FACET_SYMBOLIC_NAME, diagramClass, null);
    }

    /**
     * プロジェクト内に同じ種類の図が存在する場合に、新規作成する図かどうか確認します。
     * 
     * @param diagram
     *            図
     * @return true 新規作成する図の場合 / false それ以外の場合
     */
    private boolean isCreateNewDiagram(Diagram diagram) {
        return diagram instanceof STAMPControlLoopDiagram
                || diagram instanceof HCFTable;
    }

    /**
     * 表で表現されるDiagramを開きます。
     * 
     * @param diagram
     *            開くDiagram
     */
    private void openMatrixDiagram(Diagram diagram) {
        OpenTableEditorCommand command = new OpenTableEditorCommand();
        command.setElement(diagram);
        CommandManager.instance().callCommand(command);
    }

    /**
     * 図を開きます。
     * 
     * @param diagram
     *            開くDiagram
     */
    private void openDiagram(Diagram diagram) {
        OpenDiagramEditorCommand command = new OpenDiagramEditorCommand();
        DiagramViewInfo viewInfo = new DiagramViewInfo(diagram, new GCamera());
        command.setDiagramViewInfo(viewInfo);
        CommandManager.instance().callCommand(command);
    }

    private Color getNodeForegroundColor(Object node, boolean selected) {
        if (selected) {
            return SELECTED_COLOR;
        }
        return isPrepared(node) ? PREPARED_COLOR : NON_PREPARED_COLOR;
    }

    private boolean isPrepared(Object node) {
        if (!(node instanceof DefaultMutableTreeNode)) {
            return false;
        }
        Object uo = ((DefaultMutableTreeNode) node).getUserObject();
        return (uo instanceof TreeUserObject) ? ((TreeUserObject) uo).isPrepared() : false;
    }

    private static class DiagramNodePredicate extends STAMPNotationSwitch<Boolean> {
        @Override
        public Boolean casePreconditionTable(PreconditionTable object) {
            STPAAnalysis scope = object.getScope();
            return !scope.getPreconditions().isEmpty();
        }

        @Override
        public Boolean caseAccidentHazardSafetyConstraintTable(
                AccidentHazardSafetyConstraintTable object) {
            STPAAnalysis scope = object.getScope();
            return !scope.getAccidents().isEmpty();
        }

        @Override
        public Boolean caseComponentExtractingTable(ComponentExtractingTable object) {
            return object.getScope().getControlStructures().stream()
                    .anyMatch(cs -> !cs.getComponents().isEmpty());
        }

        @Override
        public Boolean caseUCATable(UCATable object) {
            UCAAnalysis analysis = object.getAnalysis();
            return !analysis.getControlActions().isEmpty();
        }

        @Override
        public Boolean caseCountermeasureTable(CountermeasureTable object) {
            STPAAnalysis scope = object.getScope();
            return !scope.getCountermeasures().isEmpty();
        }

        @Override
        public Boolean caseHCFTable(HCFTable object) {
            return true;
        }

        @Override
        public Boolean caseSTAMPControlStructureDiagram(STAMPControlStructureDiagram object) {
            return true;
        }

        @Override
        public Boolean caseSTAMPControlLoopDiagram(STAMPControlLoopDiagram object) {
            return true;
        }

        @Override
        public Boolean defaultCase(EObject object) {
            return false;
        }

    }

    /**
     * Treeに設定されるUserObjectのモデルクラスです。
     */
    class TreeUserObject {
        private String name;
        private String commandName;
        private Class<? extends Diagram> targetDiagramClass;
        private String toolTipText;
        private DiagramNodePredicate predicate;

        /**
         * コンストラクタ
         * 
         * @param name
         *            表示名
         */
        TreeUserObject(String name, DiagramNodePredicate predicate) {
            this(name, null, null, null, predicate);
        }

        /**
         * コンストラクタ
         * 
         * @param name
         *            表示名
         * @param toolTipText
         *            ツールチップテキスト
         */
        TreeUserObject(String name, String toolTipText) {
            this(name, null, null, toolTipText, null);
        }

        /**
         * コンストラクタ
         * 
         * @param name
         *            表示名
         * @param commandName
         *            図表作成コマンド名
         * @param targetDiagramClass
         *            探索対象図表クラス
         * @param toolTipText
         *            ツールチップテキスト
         */
        TreeUserObject(String name, String commandName, Class<? extends Diagram> targetDiagramClass,
                String toolTipText, DiagramNodePredicate predicate) {
            this.name = name;
            this.commandName = commandName;
            this.targetDiagramClass = targetDiagramClass;
            this.toolTipText = toolTipText;
            this.predicate = predicate;
        }

        @Override
        public String toString() {
            return name;
        }

        /**
         * ノードが表現するDiagramを取得します。
         * 
         * @return Diagram
         */
        public Diagram getDiagram() {
            if (targetDiagramClass == null) {
                return null;
            }

            return AnalysisProcedureView.this.getDiagram(targetDiagramClass);
        }

        /**
         * Diagramのノードかどうか確認します。
         * 
         * @return true ノードがDiagramの場合 / false ノードがDiagramではない場合
         */
        public boolean isDiagramNode() {
            return targetDiagramClass != null;
        }

        /**
         * 図表を作成します。
         */
        public void createDiagram() {
            CommandEvent event = new CommandEvent(this, 0, commandName, 0,
                    CommandEvent.RUN_TYPE_THREAD_SAME);
            if (CommandManager.instance() != null) {
                CommandManager.instance().commandPerformed(event);
            }
        }

        /**
         * ノードのツールチップテキストを取得します。
         * 
         * @return ツールチップテキスト
         */
        public String getToolTipText() {
            return toolTipText;
        }

        public boolean isPrepared() {
            if (predicate != null) {
                final Diagram diagram = getDiagram();
                return (diagram != null) && predicate.doSwitch(diagram);
            }
            return false;
        }
    }

    @Override
    public void entityStoreChanged(EntityStoreEvent e) {
        tree.updateUI();
    }
}
