package net.astah.stpa.stamp.ui.view;

import static net.astah.stpa.stamp.STAMPPackage.Literals.CONTROL_LOOP__HINT_WORD_SET;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HCF_ANALYSIS_RESULT__HINT_WORD;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HCF_HINT_WORD_SET;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyEvent;
import java.beans.PropertyVetoException;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Properties;
import java.util.ResourceBundle;

import javax.swing.DefaultListCellRenderer;
import javax.swing.JComboBox;
import javax.swing.JDesktopPane;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JLayeredPane;
import javax.swing.JPanel;
import javax.swing.WindowConstants;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;

import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.URIConverter;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;
import org.icepdf.ri.common.MyAnnotationCallback;
import org.icepdf.ri.common.SwingController;
import org.icepdf.ri.common.SwingViewBuilder;
import org.icepdf.ri.common.views.DocumentViewController;
import org.icepdf.ri.util.PropertiesManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import net.astah.emf.common.ecore.EcoreCopier;
import net.astah.emf.common.ecore.resource.DataURIHandlerImpl;
import net.astah.emf.edit.job.CommandJob;
import net.astah.emf.edit.job.Jobs;
import net.astah.emf.edit.provider.ExtraLabelVariant;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.emf.edit.provider.swing.AdapterFactoryVariantLabelProvider;
import net.astah.emf.edit.provider.swing.renderer.LabelProviderListCellRenderer;
import net.astah.golf.view.swing.UIViewPeer;
import net.astah.golf.widget.swing.SwingComponent;
import net.astah.golf.widget.swing.SwingDesktopPane;
import net.astah.golf.widget.swing.SwingIconRegistry;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jsystem.PropertyManager;
import net.astah.jomt.jutil.SwingInvoker;
import net.astah.jomt.jview.DiagramEditor;
import net.astah.jomt.jview.Editor;
import net.astah.jomt.jview.ManagementView;
import net.astah.jomt.jview.MessageManager;
import net.astah.jomt.jview.TableEditor;
import net.astah.project.Facets;
import net.astah.project.Project;
import net.astah.project.ProjectManagerAdapter;
import net.astah.stpa.stamp.ControlLoop;
import net.astah.stpa.stamp.HCFAnalysis;
import net.astah.stpa.stamp.HCFAnalysisResult;
import net.astah.stpa.stamp.HCFHintWordSet;
import net.astah.stpa.stamp.STAMPPlugin;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.notation.HCFTable;
import net.astah.stpa.stamp.notation.STAMPControlLoopDiagram;
import net.astah.stpa.stamp.ui.Messages;
import net.astah.stpa.stamp.ui.view.swing.HCFTableEditorPeer;
import net.astah.stpa.stamp.util.HCFHintWordSets;

/**
 * ヒントワードセット画像を表示するパレットフレームの本体です。
 * ヒントワードセット画像は{@link JInternalFrame}内に表示され、frameは{@link JDesktopPane}のPALETTE_LAYERに追加されます。
 * PDFの表示にはicePDFを使用しています。
 */
public class HCFHintWordSetFrame extends JInternalFrame implements ActionListener {

    private static final long serialVersionUID = 1153537152200366219L;

    private static final Logger LOGGER = LoggerFactory.getLogger(HCFHintWordSetFrame.class);

    // ヒントワードセット選択用のコンボボックス
    private JComboBox<HCFHintWordSet> comboBox;

    // PDFファイル管理用のController
    private SwingController controller;

    // Resource Set
    private final ResourceSet resourceSet;

    /**
     * ヒントワードセット画像を表示するパレットフレームを構築します。
     * 
     * @see JInternalFrame
     */
    public HCFHintWordSetFrame(final EditingDomain domain, final String title,
            final boolean resizable, final boolean closable, final boolean maximizable,
            final boolean iconifiable) {
        super(title, resizable, closable, maximizable, iconifiable);
        resourceSet = domain.getResourceSet();
        final URIConverter uriConverter = resourceSet.getURIConverter();
        DataURIHandlerImpl.install(uriConverter);
    }

    public void createHintWordSetFrame(final EList<HCFHintWordSet> hintWordSets,
            final ControlLoop controlLoop) {
        final JDesktopPane desktopPane = getViewDesktopPane();

        try {
            // ヒントワードセット選択ComboBoxの作成
            final JPanel comboPanel = createComboBoxPanel(controlLoop, hintWordSets);
            add(comboPanel, BorderLayout.NORTH);

            // PDF表示部分のJPanelを作成する
            final JPanel panel = createHintWordSetPdfPanel();
            add(panel);

            pack();
            desktopPane.add(this, JLayeredPane.PALETTE_LAYER);
            DiagramEditor.setPaletteFrame(this);
            setVisible(true);
        } catch (final IOException e) {
            LOGGER.error("Creation of HCFHintWordSet InputStream failed.", e);
        }
    }

    /**
     * パレットフレームを初期化します。
     * 
     * @throws PropertyVetoException
     */
    public void initializeFrame() throws PropertyVetoException {
        setFrameIcon(null);
        setMaximum(false);
        setBounds(300, 50, 550, 475);
        putClientProperty("JInternalFrame.isPalette", Boolean.TRUE);
        putClientProperty("JInternalFrame.frameType", "optionDialog");
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        addInternalFrameListener(new HintWordPaletteAdapter());
        addComponentListener(new FrameControlAdapter());
    }

    /**
     * パレットフレームを更新します。
     * 
     * @throws IOException
     */
    public void update() throws IOException {
        if (controller == null) {
            return;
        }
        controller.closeDocument();

        final HCFHintWordSet hintWordSet = (HCFHintWordSet) comboBox.getSelectedItem();
        final InputStream inputStream = createInputStream(hintWordSet.getImage());

        final DocumentViewController documentViewController = controller
                .getDocumentViewController();
        final MyAnnotationCallback annotationCallback = new MyAnnotationCallback(
                documentViewController);
        documentViewController.setAnnotationCallback(annotationCallback);

        controller.openDocument(inputStream, null, null);
        controller.setUtilityPaneVisible(false);
        controller.setToolBarVisible(false);
    }

    /**
     * ヒントワードセットコンボボックスを更新します
     * 
     * @param controlLoop
     *            ヒントワードセットが含まれるコントロールループ
     */
    public void updateComboBox(final ControlLoop controlLoop) {
        final STPAAnalysis analysis = getSTPAAnalysis(controlLoop);
        final EList<HCFHintWordSet> hintWordSets = HCFHintWordSets.getLanguageHintWordSets(analysis,
                controlLoop.eResource().getResourceSet());
        final boolean thisIsListening = Arrays.asList(comboBox.getActionListeners()).contains(this);
        comboBox.removeActionListener(this);
        comboBox.removeAllItems();
        for (int i = 0; i < hintWordSets.size(); i++) {
            comboBox.insertItemAt(hintWordSets.get(i), i);
        }
        HCFHintWordSet hintWordSet = controlLoop.getHintWordSet();
        comboBox.setSelectedItem(hintWordSet);
        if (thisIsListening) {
            comboBox.addActionListener(this);
        }
    }

    private JPanel createComboBoxPanel(final ControlLoop controlLoop,
            final EList<HCFHintWordSet> hintWordSets) {
        if (comboBox == null) {
            comboBox = new JComboBox<>();
        }

        comboBox.removeActionListener(this);
        comboBox.removeAllItems();
        for (final HCFHintWordSet hintWordSet : hintWordSets) {
            comboBox.addItem(hintWordSet);
        }

        comboBox.setRenderer(getComboBoxCellRenderer(controlLoop));

        final HCFHintWordSet controlLoopHintWordSet = controlLoop.getHintWordSet();
        // プロジェクトを開いた時に、設定されているヒントワードセットを選択する
        comboBox.setSelectedItem(controlLoopHintWordSet);

        final JPanel comboPanel = new JPanel();
        comboPanel.setLayout(new BorderLayout());
        comboBox.setPreferredSize(new Dimension(290, 25));
        comboPanel.add(comboBox, BorderLayout.WEST);

        comboBox.addActionListener(this);

        return comboPanel;
    }

    private JPanel createHintWordSetPdfPanel() throws IOException {
        if (controller == null) {
            controller = new SwingController() {
                @Override
                public void keyPressed(final KeyEvent e) {
                    // キー操作を無効にする
                }
            };
        } else {
            controller.closeDocument();
            controller.dispose();
        }

        Properties p = new Properties();
        p.setProperty("application.showLocalStorageDialogs", "false");
        p.setProperty("application.datadir",
                PropertyManager.getJudeResourceBundle().getString(PropertyManager.USER_CONFIG_DIR)
                        + File.separator + "icepdf_viewer");
        p.setProperty(PropertiesManager.PROPERTY_SHOW_STATUSBAR, "false");
        final PropertiesManager properties = new PropertiesManager(System.getProperties(), p,
                ResourceBundle.getBundle(PropertiesManager.DEFAULT_MESSAGE_BUNDLE));
        final SwingViewBuilder factory = new SwingViewBuilder(controller, properties);
        final JPanel panel = factory.buildViewerPanel();

        update();

        panel.setPreferredSize(new Dimension(SwingIconRegistry.scaleValue(500),
                SwingIconRegistry.scaleValue(500)));
        controller.setZoom((float) SwingIconRegistry.getDefaultScale());
        return panel;
    }

    private InputStream createInputStream(final String strURI) throws IOException {
        final URI uri = URI.createURI(strURI);
        final URIConverter uriConverter = resourceSet.getURIConverter();
        return uriConverter.createInputStream(uri, null);
    }

    private JDesktopPane getViewDesktopPane() {
        final ManagementView managementView = JomtGlobal.viewManager.getManagementView();
        final SwingDesktopPane swingDesktopPane = (SwingDesktopPane) managementView
                .getDesktopPane();
        return (JDesktopPane) swingDesktopPane.getComponent();
    }

    private DefaultListCellRenderer getComboBoxCellRenderer(final Notifier notifier) {
        final STPAAnalysis analysis = getSTPAAnalysis(notifier);
        if (analysis == null) {
            return null;
        }
        final LabelProviderListCellRenderer renderer = new LabelProviderListCellRenderer(
                new AdapterFactoryVariantLabelProvider(
                        ItemProviders.getAdapterFactory(analysis),
                        ExtraLabelVariant.MODEL_LIST));
        return renderer;
    }

    private STPAAnalysis getSTPAAnalysis(final Notifier notifier) {
        final Project project = ProjectManagerAdapter.getRootProject(notifier);
        return Facets.getFirstContent(project, STAMPPlugin.FACET_SYMBOLIC_NAME, STPAAnalysis.class,
                null);
    }

    @Override
    public void actionPerformed(final ActionEvent e) {
        if (e.getSource() == comboBox) {
            SwingInvoker.runLaterInEDT(() -> {
                hintWordSetChanged();
            });
        }
    }

    private void hintWordSetChanged() {
        final boolean thisIsListening = Arrays.asList(comboBox.getActionListeners()).contains(this);
        try {
            comboBox.removeActionListener(this);
            final ControlLoop controlLoop = getCurrentControlLoop();
            if (controlLoop == null) {
                return;
            }
            if (!(comboBox.getSelectedItem() instanceof HCFHintWordSet)) {
                return;
            }
            final HCFHintWordSet selectedItem = (HCFHintWordSet) comboBox.getSelectedItem();

            final STPAAnalysis stpaAnalysis = getSTPAAnalysis(selectedItem);
            final HCFAnalysis hcfAnalysis = stpaAnalysis.getHcfAnalysis();
            if (selectedItem != controlLoop.getHintWordSet()
                    && hcfAnalysis.getResults().stream().anyMatch(result -> {
                        return result.eIsSet(HCF_ANALYSIS_RESULT__HINT_WORD)
                                && isClearHintWord(controlLoop, result);
                    })) {
                final int yesno = MessageManager.displayYesNoDialog("",
                        Messages.hcf_hintword_set_frame_ask_change_hint_word_set_label);
                if (yesno != MessageManager.YES) {
                    comboBox.setSelectedItem(controlLoop.getHintWordSet());
                    comboBox.updateUI();
                    return;
                }
            }

            executeHintWordSetSetCommand(controlLoop, selectedItem);
        } finally {
            if (thisIsListening) {
                comboBox.addActionListener(this);
            }
        }
    }

    private boolean isClearHintWord(final ControlLoop controlLoop, final HCFAnalysisResult result) {
        // ヒントワードセットを変更した時にクリアするヒントワードを
        // 対象のControlLoopが参照しているものだけに限定する
        return controlLoop.getTargetLink().getActions()
                .contains(result.getUnsafeControlAction().getControlAction());
    }

    private void executeHintWordSetSetCommand(final ControlLoop controlLoop,
            final HCFHintWordSet hintWordSet) {
        final TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(controlLoop);
        final CompoundCommand command = new CompoundCommand();
        // ヒントワードセットがプロジェクトに存在しない場合には、取り込む
        final STPAAnalysis stpaAnalysis = getSTPAAnalysis(controlLoop);
        if (stpaAnalysis == null) {
            return;
        }
        final HCFHintWordSet hintWordSetInProject;
        if (!HCFHintWordSets.isHCFHintWordSetPresent(hintWordSet, stpaAnalysis)) {
            final EcoreCopier copier = new EcoreCopier();
            final EObject copyObject = copier.copy(hintWordSet);
            command.append(AddCommand.create(domain, stpaAnalysis, HCF_HINT_WORD_SET, copyObject));
            hintWordSetInProject = (HCFHintWordSet) copyObject;
            comboBox.insertItemAt(hintWordSetInProject, stpaAnalysis.getHCFHintWordSets().size());
        } else {
            hintWordSetInProject = HCFHintWordSets.getHintWordSetInProject(hintWordSet,
                    stpaAnalysis);
        }
        if (hintWordSetInProject != controlLoop.getHintWordSet()) {
            for (final HCFAnalysisResult result : stpaAnalysis.getHcfAnalysis().getResults()) {
                if (isClearHintWord(controlLoop, result)) {
                    command.append(SetCommand.create(domain, result, HCF_ANALYSIS_RESULT__HINT_WORD,
                            SetCommand.UNSET_VALUE));
                }
            }
            command.append(SetCommand.create(domain, controlLoop, CONTROL_LOOP__HINT_WORD_SET,
                    hintWordSetInProject));
        }
        comboBox.setSelectedItem(hintWordSetInProject);

        // コマンドを非同期ジョブとして実行する
        final CommandJob job = new CommandJob(domain, command);
        Jobs.toCompletableFuture(job, true).thenRun(() -> {
            final HCFHintWordSetFrame frame = (HCFHintWordSetFrame) DiagramEditor.getPaletteframe();
            if (frame == null) {
                return;
            }
            try {
                comboBox.updateUI();
                frame.update();
            } catch (final IOException e) {
                LOGGER.error("Creation of HCFHintWordSet ComboBox failed.", e);
            }
            // HCF表がアクティブになっている場合に、ヒントワードセットリストの更新処理を行う
            final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
            if (editor instanceof TableEditor) {
                final UIViewPeer editorPeer = editor.getUiViewPeer();
                if (editorPeer instanceof HCFTableEditorPeer) {
                    ((HCFTableEditorPeer) editorPeer).tableEditorActivated(null);
                }
            }
        });
        job.schedule();
    }

    /**
     * カレントのエディタに紐づくコントロールループを取得します。
     * <p>
     * カレントエディタが無い場合やコントロールループに紐づいていない場合は{@code null}を返します。
     * </p>
     * 
     * @return コントロールループ
     */
    public static ControlLoop getCurrentControlLoop() {
        final EObject diagram;
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        if (editor instanceof DiagramEditor) {
            diagram = ((DiagramEditor) editor).getDiagram();
        } else if (editor instanceof TableEditor) {
            diagram = ((TableEditor) editor).getModel();
        } else {
            return null;
        }

        if (diagram instanceof STAMPControlLoopDiagram) {
            return ((STAMPControlLoopDiagram) diagram).getControlLoop();
        } else if (diagram instanceof HCFTable) {
            return ((HCFTable) diagram).getControlLoop();
        }
        return null;
    }

    private class HintWordPaletteAdapter extends InternalFrameAdapter {
        @Override
        public void internalFrameClosing(final InternalFrameEvent e) {
            e.getInternalFrame().setVisible(false);

            // フォーカスを持つComponentが無くなるので、カレントエディタにフォーカスを設定する
            // 実装は SwingManagementViewPeer.EditorTabListener.stateChanged(ChangeEvent) を参照した
            final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
            if (editor == null) {
                return;
            }
            final SwingComponent sc = (SwingComponent) editor.getComponent();
            if (sc == null) {
                return;
            }
            final Component component = sc.getComponent();
            if (component instanceof JFrame) {
                ((JFrame) component).toFront();
            } else if (component instanceof JInternalFrame) {
                ((JInternalFrame) component).moveToFront();
                try {
                    ((JInternalFrame) component).setSelected(true);
                } catch (final PropertyVetoException ex) {
                    LOGGER.error("JInternalFrame.setSelected() failed.", ex);
                }
            }
        }
    }

    private class FrameControlAdapter extends ComponentAdapter {
        @Override
        public void componentMoved(final ComponentEvent e) {
            // 枠外に移動しようとした際はframeの位置を調整する
            final Component component = e.getComponent();
            final Point point = component.getLocation();

            if (point.getX() < 0) {
                point.x = 0;
            }
            if (point.getY() < 0) {
                point.y = 0;
            }
            component.setLocation(point);
        }
    }
}
