package net.astah.stpa.stamp.ui.view;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSeparator;

import net.astah.emf.edit.provider.ExtraLabelVariant;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.emf.edit.provider.swing.AdapterFactoryVariantLabelProvider;
import net.astah.emf.edit.provider.swing.renderer.LabelProviderListCellRenderer;
import net.astah.golf.widget.swing.ILabelProvider;
import net.astah.jomt.jsystem.PropertyManager;
import net.astah.jomt.jview.KeyDialog;
import net.astah.stpa.stamp.NamedElement;

/**
 * {@link NamedElement} を選択するダイアログを提供します。
 */
public class SelectNamedElementDialog extends KeyDialog implements ActionListener {

    private static final long serialVersionUID = -1;
    private int resultState = CANCEL;
    private NamedElement element;
    private List<NamedElement> elementList;

    /** OK押下時の結果コード定義 */
    public static final int OK = 1;

    /** CANCEL押下時の結果コード定義 */
    public static final int CANCEL = 2;

    private JComboBox<NamedElement> comboBox;
    private String description = null;

    /**
     * ダイアログを構築します.
     * 
     * @param owner
     *            ダイアログが表示される {@code Frame}
     * @param elementList
     *            選択肢項目群
     * @param selectedElement
     *            初期選択項目
     */
    public SelectNamedElementDialog(Frame owner, List<NamedElement> elementList,
            NamedElement selectedElement) {
        super(owner, true);

        this.elementList = elementList;
        setBounds((int) owner.getBounds().getCenterX(), (int) owner.getBounds().getCenterY(), 400,
                100);
        Container contentPane = getContentPane();
        contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.Y_AXIS));
        comboBox = createJCombobox();
        if (selectedElement != null) {
            comboBox.setSelectedItem(selectedElement);
        }

        contentPane.add(comboBox);
        contentPane.add(createOkPanel());
    }

    private Component createOkPanel() {
        JPanel okPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        JButton okButton = new JButton(PropertyManager.getProjectViewResourceBundle()
                .getString("projectview.button.ok.label"));
        okButton.setActionCommand("OK");
        okButton.addActionListener(this);
        JButton cancelButton = new JButton(PropertyManager.getProjectViewResourceBundle()
                .getString("projectview.button.cancel.label"));
        cancelButton.setActionCommand("CANCEL");
        cancelButton.addActionListener(this);
        if (PropertyManager.isMacOS()) {
            okPanel.add(cancelButton);
            okPanel.add(okButton);
        } else {
            okPanel.add(okButton);
            okPanel.add(cancelButton);
        }

        JPanel okBasePanel = new JPanel(new BorderLayout());
        okBasePanel.add(new JSeparator(), BorderLayout.CENTER);
        okBasePanel.add(okPanel, BorderLayout.SOUTH);
        return okBasePanel;
    }

    private JComboBox<NamedElement> createJCombobox() {
        JComboBox<NamedElement> cmb = new JComboBox<NamedElement>(
                elementList.toArray(new NamedElement[] {}));
        cmb.setRenderer(new LabelProviderListCellRenderer(createLabelProvider()));
        cmb.setEditable(false);
        return cmb;
    }

    /**
     * 選択用コンボボックスに適用するLabelProvider を作成します。
     * 
     * @return 適用するLabelProvider
     */
    protected ILabelProvider createLabelProvider() {
        return new AdapterFactoryVariantLabelProvider(
                ItemProviders.getAdapterFactory(elementList.get(0).eContainer()),
                ExtraLabelVariant.MODEL_LIST);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        String action = e.getActionCommand();
        if (action.equals("OK")) {
            resultState = OK;
            element = (NamedElement) comboBox.getSelectedItem();
        } else if (action.equals("CANCEL")) {
            resultState = CANCEL;
        }
        setVisible(false);
    }

    /**
     * ダイアログの結果コードを返します.
     * 
     * @return {@link #OK} または {@link #CANCEL}
     */
    public int getResultState() {
        return resultState;
    }

    /**
     * OK押下時に選択されていた項目を返します.
     * 
     * @return 被選択項目
     */
    public NamedElement getSelectedElement() {
        return element;
    }

    /**
     * 説明文言を設定します.
     * 
     * @param description
     *            説明文言
     *            <p>
     *            非nullの場合に表示します.nullの場合は表示域を消去します.
     *            </p>
     */
    public void setDescription(final String description) {
        final Container contentPane = getContentPane();
        if (this.description != null) {
            contentPane.remove(0);
            pack();
        }

        if (description != null) {
            contentPane.add(createDescriptionPanel(description), 0);
            pack();
        }

        this.description = description;
    }

    private Component createDescriptionPanel(final String description) {
        final JPanel panel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        final JLabel label = new JLabel();
        label.setText(description);
        panel.add(label);
        return panel;
    }

}
