package net.astah.stpa.stamp.ui.view.proptab;

import java.awt.BorderLayout;

import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;

import org.eclipse.emf.common.notify.AdapterFactory;
import org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain;
import org.eclipse.emf.edit.domain.EditingDomain;

import net.astah.emf.edit.provider.ExtraLabelVariant;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.emf.edit.provider.swing.AdapterFactoryContentProvider;
import net.astah.emf.edit.provider.swing.AdapterFactoryVariantLabelProvider;
import net.astah.emf.edit.provider.swing.JTableBuilder;
import net.astah.golf.view.swing.TableColumnProvider;
import net.astah.golf.widget.swing.ILabelProvider;
import net.astah.golf.widget.swing.LabelProvider;
import net.astah.jomt.jview.proptab.PropertySourceTab;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.Signal;
import net.astah.stpa.stamp.ui.Messages;

/**
 * コンポーネントに関係するコントロールアクション、フィードバックを表示するプロパティタブの汎用実装です。
 */
public abstract class STAMPComponentSignalTab extends PropertySourceTab {

    private static final long serialVersionUID = 237840439974158711L;

    JTable table;
    JScrollPane tablePanel;

    @Override
    protected void initPanel() {
        super.initPanel();
        if (baseItemsPanel == null) {
            return;
        }
        // テーブル作成
        table = createTable();
        if (tablePanel == null) {
            tablePanel = new JScrollPane();
        }
        tablePanel.setViewportView(table);
        add(tablePanel, BorderLayout.CENTER);
    }

    private JTable createTable() {
        EditingDomain domain = AdapterFactoryEditingDomain.getEditingDomainFor(getModel());
        AdapterFactory adapterFactory = ItemProviders.getAdapterFactory(domain);
        AdapterFactoryContentProvider contentProvider = createContentProvider(domain,
                adapterFactory);

        JTable table = JTableBuilder.builder(domain).adapterFactory(adapterFactory)
                .contentProvider(contentProvider).input(getInput())
                .addColumn(TableColumnProvider.label(
                        Messages.projectview_component_control_action_tab_io, getIOLabelProvider()))
                .addColumn(TableColumnProvider.label(
                        Messages.projectview_component_control_action_tab_name,
                        new AdapterFactoryVariantLabelProvider(adapterFactory,
                                ExtraLabelVariant.MODEL_TREE)))
                .addColumn(TableColumnProvider.label(
                        Messages.projectview_component_control_action_tab_connectto,
                        getConnectingComponentLabelProvider()))
                .build();
        table.setDragEnabled(false);
        table.getTableHeader().setReorderingAllowed(false);
        table.getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        return table;
    }

    /**
     * テーブルに表示するオブジェクトを供給する {@link AdapterFactoryContentProvider} を構築します。
     * 
     * @param domain
     *            EditingDomain
     * @param adapterFactory
     *            AdapterFactory
     * @return 構築した {@code AdapterFactoryContentProvider}
     */
    protected abstract AdapterFactoryContentProvider createContentProvider(EditingDomain domain,
            AdapterFactory adapterFactory);

    private ILabelProvider getConnectingComponentLabelProvider() {
        return new LabelProvider() {
            @Override
            public String getText(Object element) {
                Link link = ((Signal) element).getLink();
                if (isOutput(link)) {
                    return link.getTarget().getName();
                } else {
                    return link.getSource().getName();
                }
            }
        };
    }

    private ILabelProvider getIOLabelProvider() {
        return new LabelProvider() {
            @Override
            public String getText(Object element) {
                if (element instanceof Signal) {
                    Link link = ((Signal) element).getLink();
                    if (isOutput(link)) {
                        if (isInput(link)) {
                            return Messages.projectview_component_control_action_tab_both;
                        }
                        return Messages.projectview_component_control_action_tab_output;
                    } else if (isInput(link)) {
                        return Messages.projectview_component_control_action_tab_input;
                    }
                }
                return "";
            }
        };
    }

    private boolean isOutput(Link link) {
        return link.getSource() == getModel();
    }

    private boolean isInput(Link link) {
        return link.getTarget() == getModel();
    }

    @Override
    public void setInitialized(boolean b) {
        super.setInitialized(b);
        if (!b) {
            if (table != null) {
                JTableBuilder.dispose(table);
                tablePanel.remove(table);
                table = null;
            }
        }
    }

}
