package net.astah.stpa.stamp.ui.view.proptab;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Arrays;
import java.util.Collections;

import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.tree.TreePath;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.notify.AdapterFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.emf.edit.command.DestroyCommand;
import net.astah.emf.edit.job.CommandJob;
import net.astah.emf.edit.job.Jobs;
import net.astah.emf.edit.provider.ExtraLabelVariant;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.emf.edit.provider.swing.AdapterFactoryVariantLabelProvider;
import net.astah.emf.edit.provider.swing.DecoratorJTreeBuilder;
import net.astah.emf.edit.provider.swing.EditingDomainTreeTransferHandler;
import net.astah.jomt.jview.JomtTree;
import net.astah.jomt.jview.swing.TreeNodeExpander;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.ProcessModel;
import net.astah.stpa.stamp.ProcessValue;
import net.astah.stpa.stamp.ProcessVariable;
import net.astah.stpa.stamp.STAMPFactory;
import net.astah.stpa.stamp.STAMPPackage;
import net.astah.stpa.stamp.ui.Messages;

/**
 * コンポーネント に関係するプロセスモデルを表示するプロパティタブの実装です。
 */
public class STAMPProcessModelTab extends STAMPModelTab implements ActionListener {
    private static final long serialVersionUID = 6143077688312421486L;
    private JPanel processModelPanel;
    private JButton createProcessModelButton;
    private JTree processModelTree;
    private final JMenuItem deleteProcessModelMenu;
    private final JMenuItem addVariableMenu;
    private final JMenuItem deleteVariableMenu;
    private final JMenuItem addValueMenu;
    private final JMenuItem deleteValueMenu;

    public STAMPProcessModelTab() {
        createProcessModelButton = new JButton(
                Messages.stamp_process_model_tab_create_process_model_label);
        createProcessModelButton.addActionListener(this);
        deleteProcessModelMenu = new JMenuItem(
                Messages.stamp_process_model_tab_delete_process_model_label);
        deleteProcessModelMenu.addActionListener(this);
        addVariableMenu = new JMenuItem(Messages.stamp_process_model_tab_add_variable_label);
        addVariableMenu.addActionListener(this);
        deleteVariableMenu = new JMenuItem(Messages.stamp_process_model_tab_delete_variable_label);
        deleteVariableMenu.addActionListener(this);
        addValueMenu = new JMenuItem(Messages.stamp_process_model_tab_add_value_label);
        addValueMenu.addActionListener(this);
        deleteValueMenu = new JMenuItem(Messages.stamp_process_model_tab_delete_value_label);
        deleteValueMenu.addActionListener(this);
    }

    @Override
    public void refresh(Object element) {
        super.refresh(element);
        if (!(element instanceof Component)) {
            return;
        }
        Component model = (Component) element;

        if (Arrays.asList(processModelPanel.getComponents())
                .contains(createProcessModelButton)) {
            if (model.getProcessModel() != null) {
                initPanel();
            }
        } else {
            if (model.getProcessModel() == null) {
                initPanel();
            }
        }
        updateUI();
    }

    @Override
    protected void initPanel() {
        super.initPanel();
        removeAll();
        processModelPanel = new JPanel();
        processModelPanel.setLayout(new BorderLayout());
        if (model instanceof Component) {
            if (!hasProcessModel(model)) {
                processModelPanel.add(createProcessModelButton, BorderLayout.NORTH);
            } else {
                ProcessModel processModel = ((Component) model).getProcessModel();
                TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(processModel);

                processModelTree = new JomtTree();
                AdapterFactory adapterFactory = ItemProviders.getAdapterFactory(domain);
                DecoratorJTreeBuilder.builder(domain)
                        .labelProvider(new AdapterFactoryVariantLabelProvider(adapterFactory,
                                ExtraLabelVariant.NAME))
                        .transferHandler(new EditingDomainTreeTransferHandler(domain))
                        .root(processModel)
                        .editable(true)
                        .build(processModelTree);
                processModelTree.setAutoscrolls(true);
                new TreeNodeExpander(processModelTree);

                processModelTree.addMouseListener(new MouseAdapter() {
                    @Override
                    public void mouseClicked(MouseEvent e) {
                        if (SwingUtilities.isRightMouseButton(e)) {
                            TreePath path = processModelTree.getClosestPathForLocation(e.getX(),
                                    e.getY());
                            processModelTree.setSelectionPath(path);
                            Object selectingObject = path.getLastPathComponent();
                            JPopupMenu popupMenu = new JPopupMenu();
                            if (selectingObject instanceof ProcessModel) {
                                popupMenu.add(addVariableMenu);
                                popupMenu.add(deleteProcessModelMenu);
                            } else if (selectingObject instanceof ProcessVariable) {
                                popupMenu.add(addVariableMenu);
                                popupMenu.add(deleteVariableMenu);
                                popupMenu.add(addValueMenu);

                            } else if (selectingObject instanceof ProcessValue) {
                                popupMenu.add(addValueMenu);
                                popupMenu.add(deleteValueMenu);
                            } else {
                                return;
                            }
                            popupMenu.show(e.getComponent(), e.getX(), e.getY());
                            e.consume();
                        }
                    }

                });
                final JScrollPane scrollPane = new JScrollPane(processModelTree);
                processModelPanel.add(scrollPane, BorderLayout.CENTER);
            }
        }
        processModelPanel.updateUI();
        add(processModelPanel);
    }

    @Override
    void initPanelRows() {
    }

    private boolean hasProcessModel(EObject model) {
        return model.eIsSet(STAMPPackage.Literals.COMPONENT__PROCESS_MODEL);
    }

    @Override
    public String getTabName() {
        return Messages.projectview_process_model_tab_title;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        Command command = null;
        TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(model);
        if (e.getSource() == createProcessModelButton) {
            command = CreateChildCommand.create(domain, model,
                    new CommandParameter(model,
                            STAMPPackage.Literals.COMPONENT__PROCESS_MODEL,
                            STAMPFactory.eINSTANCE.createProcessModel()),
                    Collections.singleton(model));
        } else {
            Object selectedModel = processModelTree.getSelectionPath().getLastPathComponent();
            if (e.getSource() == addVariableMenu) {
                final ProcessVariable processVariable = STAMPFactory.eINSTANCE
                        .createProcessVariable();
                Object owner = null;
                if (selectedModel instanceof ProcessModel) {
                    owner = selectedModel;
                } else if (selectedModel instanceof ProcessVariable) {
                    owner = ((ProcessVariable) selectedModel).eContainer();
                } else {
                    return;
                }
                command = CreateChildCommand.create(domain, owner,
                        new CommandParameter(owner,
                                STAMPPackage.Literals.PROCESS_MODEL__VARIABLE, processVariable),
                        Collections.singleton(owner));
            } else if (e.getSource() == addValueMenu) {
                final ProcessValue processValue = STAMPFactory.eINSTANCE.createProcessValue();
                Object owner = null;
                if (selectedModel instanceof ProcessVariable) {
                    owner = selectedModel;
                } else if (selectedModel instanceof ProcessValue) {
                    owner = ((ProcessValue) selectedModel).eContainer();
                } else {
                    return;
                }
                command = CreateChildCommand.create(domain, owner,
                        new CommandParameter(owner,
                                STAMPPackage.Literals.PROCESS_VARIABLE__VALUE, processValue),
                        Collections.singleton(owner));
            } else if (e.getSource() == deleteProcessModelMenu) {
                command = DestroyCommand.create(domain, selectedModel);
            } else if (e.getSource() == deleteVariableMenu) {
                command = DestroyCommand.create(domain, selectedModel);
            } else if (e.getSource() == deleteValueMenu) {
                command = DestroyCommand.create(domain, selectedModel);
            } else {
                return;
            }
        }
        CommandJob job = new CommandJob(domain, command);
        Jobs.toCompletableFuture(job, true);
        job.schedule();
    }

}
