package net.astah.stpa.stamp.ui.view.swing;

import static java.util.Arrays.asList;
import static java.util.Collections.singleton;
import static net.astah.stpa.stamp.STAMPPackage.Literals.ACCIDENT;
import static net.astah.stpa.stamp.STAMPPackage.Literals.ACCIDENT__HAZARD;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD__SAFETY_CONSTRAINT;
import static net.astah.stpa.stamp.STAMPPackage.Literals.IDENTIFIED_ELEMENT__DESCRIPTION;
import static net.astah.stpa.stamp.STAMPPackage.Literals.IDENTIFIED_ELEMENT__ID;
import static net.astah.stpa.stamp.STAMPPackage.Literals.STPA_ANALYSIS__ACCIDENT;
import static org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain.getEditingDomainFor;

import java.util.ArrayList;
import java.util.List;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.notify.AdapterFactory;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.edit.provider.IItemProviderDecorator;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.dd.di.Diagram;
import net.astah.emf.common.notify.EFeaturePath;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.golf.widget.CJComponent;
import net.astah.golf.widget.IListSelection;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.ITableSelection;
import net.astah.golf.widget.TableSelection;
import net.astah.golf.widget.swing.Selections;
import net.astah.notation.TableView;
import net.astah.stpa.stamp.Accident;
import net.astah.stpa.stamp.STAMPFactory;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.notation.AccidentHazardSafetyConstraintTable;
import net.astah.stpa.stamp.ui.Messages;
import net.astah.stpa.stamp.ui.menu.handler.AddAccidentHandler;
import net.astah.stpa.stamp.ui.menu.handler.AddHazardHandler;
import net.astah.stpa.stamp.ui.menu.handler.AddSafetyConstraintHandler;
import net.astah.stpa.stamp.ui.menu.handler.DeleteAccidentHandler;
import net.astah.stpa.stamp.ui.menu.handler.DeleteHazardHandler;
import net.astah.stpa.stamp.ui.menu.handler.DeleteSafetyConstraintHandler;
import net.astah.stpa.stamp.ui.menu.handler.SelectHazardHandler;
import net.astah.stpa.stamp.ui.menu.handler.SelectSafetyConstraintHandler;
import net.astah.table.DynamicDataset;
import net.astah.table.Element;
import net.astah.table.ElementSet;
import net.astah.table.FixedDataset;
import net.astah.table.Grid;
import net.astah.table.JoinReference;
import net.astah.table.JoinType;
import net.astah.table.ObjectDataQuery;
import net.astah.table.ObjectElement;
import net.astah.table.PropertyCellDescriptor;
import net.astah.table.Table;
import net.astah.table.TableFactory;
import net.astah.table.Tuple;
import net.astah.table.edit.provider.GenericTableAdapterFactory;
import net.astah.table.edit.provider.GridItemProviderDecorator;
import net.astah.ui.menu.annotation.Context;

/**
 * アクシデントハザード安全制約表のTableEditorPeer 実装です。
 */
@Context(id = "net.astah.stpa.stamp.notation.AccidentHazardSafetyConstraintTable",
        classes = { // classesの記述順でメニューを表示する
                AddAccidentHandler.class,
                DeleteAccidentHandler.class,
                AddHazardHandler.class,
                SelectHazardHandler.class,
                DeleteHazardHandler.class,
                AddSafetyConstraintHandler.class,
                SelectSafetyConstraintHandler.class,
                DeleteSafetyConstraintHandler.class
        })
public class AccidentHazardSafetyConstraintTableEditorPeer extends STPATableEditorPeer {
    public final int ACCIDENT_INDEX = 0;
    public final int HAZARD_INDEX = 1;
    public final int SAFETY_CONSTRAINT_INDEX = 2;

    @Override
    public void initialize(Diagram diagram, CJComponent parent) {
        super.initialize(diagram, parent);
        JTable table = (JTable) getTable().getComponent();
        table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        table.setCellSelectionEnabled(true);
    }

    @Override
    protected float[] getColumnsWidthRatio() {
        return new float[] { 1 / 12f, 3 / 12f, 1 / 12f, 3 / 12f, 1 / 12f, 3 / 12f };
    }

    @Override
    protected Table prepareTableModel(final TableView model) {
        final STPAAnalysis analysis = ((AccidentHazardSafetyConstraintTable) model).getScope();

        final ObjectDataQuery row0 = TableFactory.eINSTANCE.createObjectDataQuery();
        row0.getObjects().add(analysis);
        row0.setMetaclass(ACCIDENT);
        row0.setTraversal(true);

        final JoinReference row1 = TableFactory.eINSTANCE.createJoinReference();
        applyFeaturePath(row1, EFeaturePath.build(ACCIDENT__HAZARD));
        row1.setJoinType(JoinType.OUTER);
        row1.setOrigin(row0);

        final JoinReference row2 = TableFactory.eINSTANCE.createJoinReference();
        applyFeaturePath(row2, EFeaturePath.build(HAZARD__SAFETY_CONSTRAINT));
        row2.setJoinType(JoinType.OUTER);
        row2.setOrigin(row1);

        final DynamicDataset row = TableFactory.eINSTANCE.createDynamicDataset();
        row.getQueries().addAll(asList(row0, row1, row2));

        final ObjectElement column0 = TableFactory.eINSTANCE.createObjectElement();
        column0.setHeader(TableFactory.eINSTANCE.createFixedCellDescriptor(
                Messages.accident_hazard_safety_constraint_table_accident_id_label));
        final PropertyCellDescriptor column0Cell = TableFactory.eINSTANCE
                .createPropertyCellDescriptor();
        column0Cell.setOrigin(row.getQueries().indexOf(row0));
        applyFeaturePath(column0Cell, EFeaturePath.build(IDENTIFIED_ELEMENT__ID));
        column0Cell.setLineWrap(true);
        column0.setCell(column0Cell);

        final ObjectElement column1 = TableFactory.eINSTANCE.createObjectElement();
        column1.setHeader(TableFactory.eINSTANCE.createFixedCellDescriptor(
                Messages.accident_hazard_safety_constraint_table_accident_label));
        final PropertyCellDescriptor column1Cell = TableFactory.eINSTANCE
                .createPropertyCellDescriptor();
        column1Cell.setOrigin(row.getQueries().indexOf(row0));
        applyFeaturePath(column1Cell, EFeaturePath.build(IDENTIFIED_ELEMENT__DESCRIPTION));
        column1Cell.setLineWrap(true);
        column1.setCell(column1Cell);

        final ObjectElement column2 = TableFactory.eINSTANCE.createObjectElement();
        column2.setHeader(TableFactory.eINSTANCE.createFixedCellDescriptor(
                Messages.accident_hazard_safety_constraint_table_hazard_id_label));
        final PropertyCellDescriptor column2Cell = TableFactory.eINSTANCE
                .createPropertyCellDescriptor();
        column2Cell.setOrigin(row.getQueries().indexOf(row1));
        applyFeaturePath(column2Cell, EFeaturePath.build(IDENTIFIED_ELEMENT__ID));
        column2Cell.setLineWrap(true);
        column2.setCell(column2Cell);

        final ObjectElement column3 = TableFactory.eINSTANCE.createObjectElement();
        column3.setHeader(TableFactory.eINSTANCE.createFixedCellDescriptor(
                Messages.accident_hazard_safety_constraint_table_hazard_label));
        final PropertyCellDescriptor column3Cell = TableFactory.eINSTANCE
                .createPropertyCellDescriptor();
        column3Cell.setOrigin(row.getQueries().indexOf(row1));
        applyFeaturePath(column3Cell, EFeaturePath.build(IDENTIFIED_ELEMENT__DESCRIPTION));
        column3Cell.setLineWrap(true);
        column3.setCell(column3Cell);

        final ObjectElement column4 = TableFactory.eINSTANCE.createObjectElement();
        column4.setHeader(TableFactory.eINSTANCE.createFixedCellDescriptor(
                Messages.accident_hazard_safety_constraint_table_safety_constraint_id_label));
        final PropertyCellDescriptor column4Cell = TableFactory.eINSTANCE
                .createPropertyCellDescriptor();
        column4Cell.setOrigin(row.getQueries().indexOf(row2));
        applyFeaturePath(column4Cell, EFeaturePath.build(IDENTIFIED_ELEMENT__ID));
        column4Cell.setLineWrap(true);
        column4.setCell(column4Cell);

        final ObjectElement column5 = TableFactory.eINSTANCE.createObjectElement();
        column5.setHeader(TableFactory.eINSTANCE.createFixedCellDescriptor(
                Messages.accident_hazard_safety_constraint_table_safety_constraint_label));
        final PropertyCellDescriptor column5Cell = TableFactory.eINSTANCE
                .createPropertyCellDescriptor();
        column5Cell.setOrigin(row.getQueries().indexOf(row2));
        applyFeaturePath(column5Cell, EFeaturePath.build(IDENTIFIED_ELEMENT__DESCRIPTION));
        column5Cell.setLineWrap(true);
        column5.setCell(column5Cell);

        final ElementSet columnSet = TableFactory.eINSTANCE.createElementSet();
        columnSet.getElements()
                .addAll(asList(column0, column1, column2, column3, column4, column5));
        final FixedDataset column = TableFactory.eINSTANCE.createFixedDataset();
        column.getElementSets().add(columnSet);

        final Table table = TableFactory.eINSTANCE.createTable();
        table.getColumns().add(column);
        table.getRows().add(row);
        return table;
    }

    /**
     * アクシデントハザード安全制約表のスコープを取得します
     * 
     * @return アクシデントハザード安全制約表のスコープ
     */
    public STPAAnalysis getScope() {
        return ((AccidentHazardSafetyConstraintTable) getDiagram()).getScope();
    }

    @Override
    protected AdapterFactory createTableAdapterFactory() {
        final STPAAnalysis scope = getScope();
        final EditingDomain domain = TransactionUtil.getEditingDomain(getDiagram());
        return new GenericTableAdapterFactory(ItemProviders.getAdapterFactory(domain)) {

            @Override
            protected IItemProviderDecorator createGridDecorator() {
                return new GridItemProviderDecorator(this) {
                    @Override
                    protected Command createPrepareRowTupleCommand(final Grid grid,
                            final Tuple tuple,
                            final EditingDomain domain) {
                        return prepareAccidentCommand(tuple);
                    }
                };
            }

            protected Command prepareAccidentCommand(final Tuple rowTuple) {
                final Accident accident = STAMPFactory.eINSTANCE.createAccident();
                final EList<Element> elements = rowTuple.getElements();
                elements.get(ACCIDENT_INDEX).setObject(accident);

                final EditingDomain domain = getEditingDomainFor(scope);
                final CommandParameter childDescription = new CommandParameter(scope,
                        STPA_ANALYSIS__ACCIDENT, accident);
                return CreateChildCommand.create(domain, scope, childDescription, singleton(scope));
            }
        };
    }

    @Override
    protected ISelection getEditorSelection(final ITableSelection selection) {
        final IListSelection columnSelection = selection.getColumnSelection();
        final int[] columnIndices = columnSelection.getIndices();
        if (columnIndices.length > 0) {
            final List<Object> elements = new ArrayList<>(selection.size());
            for (final Tuple tuple : Selections.iterable(selection, Tuple.class)) {
                for (final int columnIndex : columnIndices) {
                    switch (columnIndex) {
                    default:
                    case 0: // アクシデントID
                    case 1: // アクシデントの内容
                        elements.add(tuple.getElementObject(ACCIDENT_INDEX));
                        break;
                    case 2: // ハザードID
                    case 3: // ハザードの内容
                        elements.add(tuple.getElementObject(HAZARD_INDEX));
                        break;
                    case 4: // 安全制約ID
                    case 5: // 安全制約の内容
                        elements.add(tuple.getElementObject(SAFETY_CONSTRAINT_INDEX));
                        break;
                    }
                }
            }
            final IListSelection rowSelection = selection.getRowSelection();
            return new TableSelection(elements, rowSelection, columnSelection);
        }
        return super.getEditorSelection(selection);
    }

}
